/*
 * Minidxr2 - Linux DVD player. Mostly (C) Jack Whitham, 2001
 *
 */
/*
 * buffer.c
 *
 * Functions for reading data off the disk and passing it to the dxr2
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <ctype.h>

#include "buffer.h"
#include "player.h"
#include "device.h"
#include "parser.h"
#include "ifo_parse.h"
#include "queue.h"
#include "message.h"


static void change_type_settings ( output_plugin * out , int type ,
			int def_format , int def_freq , int def_width ,
			int def_mode ) ;

char * buf_open_files ( player_info_t * player_info , 
		char ** file_list , int num_files )
{
	int			i , j ;
	struct stat statbuf ;

	if ( num_files < 1 )
	{
		return "silly number of files\n" ;
	}

	pthread_mutex_lock ( & ( player_info->mutex ) ) ;
	pthread_mutex_lock ( & ( player_info->file_access_mutex ) ) ;

	if ( player_info -> num_files > 0 )
	{
		pthread_mutex_unlock ( & ( player_info->file_access_mutex ) ) ;
		pthread_mutex_unlock ( & ( player_info->mutex ) ) ;

		return "buf_open_files: some files are already open\n" ;
	}

	player_info -> file_ptrs = calloc ( num_files , sizeof ( mfile ) ) ;
	player_info -> num_files = num_files ;
	player_info -> total_size = 0 ;
	player_info -> current_lba = 0 ;
	player_info -> read_position = player_info -> write_position = 0 ;
	player_info -> read_position_sub = 0 ;

	for ( i = 0 ; i < num_files ; i ++ )
	{
		if (( player_info -> file_ptrs [ i ] . type =
				buf_get_type_from_name ( file_list [ i ] )) == TYPE_UNKNOWN )
		{
			for ( j = 0 ; j < i ; j ++ )
			{
				free ( player_info -> file_ptrs [ i ] . name ) ;
				close ( player_info -> file_ptrs [ i ] . fd ) ;
			}
			free ( player_info -> file_ptrs ) ;
			player_info -> num_files = 0 ;

			pthread_mutex_unlock ( & ( player_info->file_access_mutex ) ) ;
			pthread_mutex_unlock ( & ( player_info->mutex ) ) ;

			return "Specified files have unsupported types. Only VOB and MPG are supported.\n" ;
		}

		if ( ( player_info -> file_ptrs [ i ] . fd = 
				open ( file_list [ i ] , O_RDONLY ) ) < 0 )
		{
			for ( j = 0 ; j < i ; j ++ )
			{
				free ( player_info -> file_ptrs [ i ] . name ) ;
				close ( player_info -> file_ptrs [ i ] . fd ) ;
			}
			free ( player_info -> file_ptrs ) ;
			player_info -> num_files = 0 ;

			pthread_mutex_unlock ( & ( player_info->file_access_mutex ) ) ;
			pthread_mutex_unlock ( & ( player_info->mutex ) ) ;

			return "Unable to open specified files.\n" ;
		}

		player_info -> file_ptrs [ i ] . name = strdup ( file_list [ i ] ) ;
		if ( fstat ( player_info -> file_ptrs [ i ] . fd , & statbuf ) == 0 )
		{
			player_info -> file_ptrs [ i ] . size = (int)
					( statbuf . st_size / BUFFER_CHUNK_SIZE ) ;

			player_info -> file_ptrs [ i ] . on_dvd =
					buf_is_on_dvd ( player_info , file_list [ i ] ) ;

			player_info -> file_ptrs [ i ] . start_pos = 
					player_info -> total_size ;
			player_info -> total_size += 
					player_info -> file_ptrs [ i ] . size ;
			player_info -> file_ptrs [ i ] . end_pos = 
					player_info -> total_size ;
		} else {
			player_info -> file_ptrs [ i ] . start_pos = 
				player_info -> file_ptrs [ i ] . end_pos = player_info -> total_size ;
			player_info -> file_ptrs [ i ] . size = 0 ;
		}
	} 
	player_info -> read_fd = player_info -> file_ptrs [ 0 ] . fd ;
	player_info -> force_reauth = 1 ;
	player_info -> current_type = player_info -> file_ptrs [ 0 ] . type ;

	pthread_mutex_unlock ( & ( player_info->file_access_mutex ) ) ;
	pthread_mutex_unlock ( & ( player_info->mutex ) ) ;
	return NULL ;
}

int buf_get_lbas ( player_info_t * player_info , char * error , int max_chars )
{
	int			i ;
	char		buf1 [ 32 ] ;
	char		buf2 [ FNAME_LENGTH ] ;
	int			capitals ;
	char		* lastslash ;
	int			title = 0 ;
	int			num = 0 ;

	pthread_mutex_lock ( & ( player_info->mutex ) ) ;
	pthread_mutex_lock ( & ( player_info->file_access_mutex ) ) ;

	for ( i = 0 ; i < player_info -> num_files ; i ++ )
	{
		player_info -> file_ptrs [ i ] . lba = 0 ;
		if ( player_info -> file_ptrs [ i ] . on_dvd )
		{
			/* work out the title number */
			lastslash = rindex ( player_info -> file_ptrs [ i ] . name , '/' ) ;
			if ( lastslash != NULL )
			{
				strncpy ( buf1 , lastslash , 31 ) ;
				buf1 [ 31 ] = '\0' ;
				/* use the 2nd letter to determine if the path is
				 * all in capitals or not (i.e. VIDEO_TS or video_ts)
				 * Not a great way but it will work.
				 */
				capitals = isupper ( buf1 [ 1 ] ) ;
				string_uppercase ( buf1 ) ;

				if (( sscanf ( buf1 , "/VTS_%02d_%d.VOB" , & title , & num ) == 2 )
				&& ( title > 0 )
				&& ( title < 100 ))
				{
					strcpy ( buf2 , player_info -> dvd_mountpoint ) ;
					remove_trailing_slashes ( buf2 ) ;

					snprintf ( buf1 , 31 , TITLE_FILENAME , title , 1 ) ;
					if ( capitals )
					{
						string_uppercase ( buf1 ) ;
					}
					if ( ( strlen ( buf1 ) + strlen ( buf2 )) >= BUFFER_LENGTH )
					{
						pthread_mutex_unlock ( & ( player_info->file_access_mutex ) ) ;
						pthread_mutex_unlock ( & ( player_info->mutex ) ) ;
						return 0 ;
					}

					strcat ( buf2 , buf1 ) ;

					if ( ! device_get_lba ( player_info , error , max_chars ,
								buf2 ,
								& player_info -> file_ptrs [ i ] . lba ) )
					{
						pthread_mutex_unlock ( & ( player_info->file_access_mutex ) ) ;
						pthread_mutex_unlock ( & ( player_info->mutex ) ) ;
						return 0 ;
					}
				}
			}
		}
		if ( player_info -> file_ptrs [ i ] . fd 
				== player_info -> read_fd )
		{
			player_info -> current_lba = 
					player_info -> file_ptrs [ i ] . lba ;
			player_info -> current_type =
					player_info -> file_ptrs [ i ] . type ;
		}
	}

	pthread_mutex_unlock ( & ( player_info->file_access_mutex ) ) ;
	pthread_mutex_unlock ( & ( player_info->mutex ) ) ;

	return 1 ;
}

int buf_files_on_dvd ( player_info_t * player_info )
{
	int		i ;

	pthread_mutex_lock ( & ( player_info->mutex ) ) ;
	pthread_mutex_lock ( & ( player_info->file_access_mutex ) ) ;
	for ( i = 0 ; i < player_info -> num_files ; i ++ )
	{
		if ( player_info -> file_ptrs [ i ] . on_dvd )
		{
			pthread_mutex_unlock ( & ( player_info->file_access_mutex ) ) ;
			pthread_mutex_unlock ( & ( player_info->mutex ) ) ;
			return 1 ;
		}
	}
	pthread_mutex_unlock ( & ( player_info->file_access_mutex ) ) ;
	pthread_mutex_unlock ( & ( player_info->mutex ) ) ;
	return 0 ;
}

void buf_clear_buffers ( player_info_t * player_info )
{
	buffer_item * item ;

	while ( ( item = (buffer_item *) 
				queue_remove_from_queue ( player_info -> buffer )) != NULL )
	{
		assert ( queue_add_to_queue ( player_info -> standby_buffer , item ) ) ;
	}
}


char * buf_close_files ( player_info_t * player_info )
{
	int			i ;
	char		* ret = NULL ;

	pthread_mutex_lock ( & ( player_info->mutex ) ) ;
	pthread_mutex_lock ( & ( player_info->file_access_mutex ) ) ;
	for ( i = 0 ; i < player_info -> num_files ; i ++ )
	{
		free ( player_info -> file_ptrs [ i ] . name ) ;
		if ( close ( player_info -> file_ptrs [ i ] . fd ) != 0 )
		{
			ret = "Failed to close one or more files.\n" ;
		}
	}
	free ( player_info -> file_ptrs ) ;
	player_info -> total_size = player_info -> num_files = 0 ;
	player_info -> read_fd = -1 ;

	player_info -> read_position = player_info -> write_position = 0 ;
	player_info -> read_position_sub = 0 ;

	buf_clear_buffers ( player_info ) ;

	pthread_mutex_unlock ( & ( player_info->file_access_mutex ) ) ;
	pthread_mutex_unlock ( & ( player_info->mutex ) ) ;

	return ret ;
}


int pbc_fd_to_file_num ( player_info_t * player_info )
{
	int 		i ;

	for ( i = 0 ; i < player_info -> num_files ; i ++ )
	{
		if ( player_info -> read_fd == player_info -> file_ptrs [ i ] . fd )
		{
			return i ;
		}
	}
	return -1 ;
}

char * pbc_seek_next_file ( player_info_t * player_info ) 
{
	int			cur_file ;

	cur_file = pbc_fd_to_file_num ( player_info ) ;
	if ( cur_file == -1 )
	{
		/* hmm.. */
		return "Couldn't work out which file is being played.\n" ;
	}

	cur_file ++ ;
	if ( cur_file >= player_info -> num_files )
	{
		return "There is no next file.\n" ;
	}

	player_info -> read_fd = player_info -> file_ptrs [ cur_file ] . fd ;
	lseek ( player_info -> read_fd , 0 , SEEK_SET ) ;
			
	player_info -> read_position_sub = 0 ;
	player_info -> read_position = 
						player_info -> file_ptrs [ cur_file ] . start_pos ;
	player_info -> current_lba = player_info -> file_ptrs [ cur_file ] . lba ;
	player_info -> current_type =
					player_info -> file_ptrs [ cur_file ] . type ;

	return NULL ;
}

char * pbc_seek ( player_info_t * player_info , int seek )
{
	int			i ;

	for ( i = 0 ; i < player_info -> num_files ; i ++ )
	{
		if (( seek >= player_info -> file_ptrs [ i ] . start_pos )
		&& ( seek < player_info -> file_ptrs [ i ] . end_pos ))
		{
			player_info -> read_fd =
					player_info -> file_ptrs [ i ] . fd ;
			lseek ( player_info -> read_fd ,
					(off_t) ( seek - 
							player_info -> file_ptrs [ i ] . start_pos ) 
						* BUFFER_CHUNK_SIZE , SEEK_SET ) ;
					
			player_info -> read_position_sub = 0 ;
			player_info -> read_position = seek ;
			player_info -> current_type =
					player_info -> file_ptrs [ i ] . type ;
			player_info -> current_lba = 
					player_info -> file_ptrs [ i ] . lba ;
			return NULL ;
		}
	}
	return "Couldn't find that position in any of the loaded files\n" ;
}

/* This function is used by functions in control.c
 * to control the behaviour of the read/write threads.
 *
 * Messages pass the other way via the functions in message.c
 */
char * buf_playback_controller ( player_info_t * player_info , 
			int command , int * parameter )
{
	char		* err = NULL ;

	/* The Initial Command */

    pthread_mutex_lock ( & ( player_info->mutex ) ) ;

	switch ( command )
	{
		case PBC_SEEK :		
				{
					int			seek_to ;

					seek_to = (* parameter) ;
					if (( seek_to >= player_info -> total_size )
					|| ( seek_to < 0 ))
					{
						err = "Cannot seek to that position - it is off the end of the file.\n" ;
					} else {
						pthread_mutex_lock ( & ( player_info->file_access_mutex ) ) ;
						err = pbc_seek ( player_info , seek_to ) ;
						buf_clear_buffers ( player_info ) ;
						player_info -> recompute_end = 1 ;
						pthread_mutex_unlock ( & ( player_info->file_access_mutex ) ) ;
					}
				}
				break ;
		case PBC_STOP :
				player_info -> b_running = 0 ;
				break ;
		case PBC_START :
				player_info -> b_running = 1 ;
				break ;
		case PBC_GET_POS :
				(* parameter) = player_info -> write_position ;
				break ;
		case PBC_QUIT :
				player_info -> b_running = 0 ;
				player_info -> b_quit = 1 ;
				pthread_cond_broadcast 
						( & ( player_info -> buffer_quite_full ) ) ;
				break ;
		case PBC_GET_LENGTH :
				(* parameter) = player_info -> total_size ;
				break ;
		case PBC_SET_TITLE :
				player_info -> read_title = (* parameter) ;
				ifo_recompute_tables ( player_info ) ;
				player_info -> recompute_end = 1 ;
				break ;
		default :
				pthread_mutex_unlock ( & ( player_info->mutex ) ) ;
				return "Unknown pbc command\n" ;
	}
	pthread_mutex_unlock ( & ( player_info->mutex ) ) ;
	pthread_cond_broadcast ( & ( player_info -> reader_state_change ) ) ;
	return err ;
}

/* 
 * ring buffer of size BUFFER_RING_SIZE, handled
 * by the functions at the bottom of this file.
 */

void * buf_reader ( void * arg )
{
	player_info_t 	* player_info = (player_info_t *) arg ;
	buffer_item		* buffer_chunk ;
	int				pos , last_pos = -1 , until = 0 ;

	pthread_mutex_lock ( & ( player_info->mutex ) ) ;
	while ( ! player_info -> b_quit )
	{
		while (( queue_is_space_in_queue ( player_info -> buffer ))
		&& ( player_info -> b_running )
		&& ( player_info -> read_fd >= 0 )
		&& ( ! player_info -> b_quit ))
		{
			buffer_chunk = queue_remove_from_queue ( player_info -> standby_buffer ) ;

			assert ( buffer_chunk != NULL ) ;

			pthread_mutex_unlock ( & ( player_info->mutex ) ) ;

			pthread_mutex_lock ( & ( player_info->file_access_mutex ) ) ;
			buffer_chunk -> size = read ( player_info -> read_fd , 
							buffer_chunk -> data , BUFFER_CHUNK_SIZE ) ;
			buffer_chunk -> source_pos = player_info -> read_position ;
			buffer_chunk -> source_lba = player_info -> current_lba ;
			buffer_chunk -> source_type = player_info -> current_type ;

			/* Move to the next file? */
			if (  buffer_chunk -> size < BUFFER_CHUNK_SIZE )
			{
				pthread_mutex_lock ( & ( player_info->mutex ) ) ;
				if ( pbc_seek_next_file ( player_info ) != NULL )
				{
					/* stop playback */
					(void) pbc_seek ( player_info , 0 ) ;
					player_info -> b_running = 0 ;
					msg_send_message ( player_info ,
									MSG_END_OF_FILE , NULL ) ;
				}
				pthread_mutex_unlock ( & ( player_info->mutex ) ) ;
			}
			pthread_mutex_unlock ( & ( player_info->file_access_mutex ) ) ;


			pthread_mutex_lock ( & ( player_info->mutex ) ) ;
			/* add data to buffer */
			assert ( queue_add_to_queue ( player_info -> buffer , buffer_chunk ) ) ;

			/* increment read position */
			if ( buffer_chunk -> size < BUFFER_CHUNK_SIZE )
			{
				player_info -> read_position_sub += buffer_chunk -> size ;
				if ( player_info -> read_position_sub >= BUFFER_CHUNK_SIZE )
				{
					player_info -> read_position += 
							player_info -> read_position_sub >> BUFFER_CHUNK_SHIFT ;
					player_info -> read_position_sub &= BUFFER_CHUNK_BITMASK ;
					player_info -> read_position ++ ;
				}
			} else {
				player_info -> read_position ++ ;
			}

			if ( player_info -> read_title > 0 )
			{
				if (( player_info -> read_position >= until )
				|| ( player_info -> read_position < last_pos )
				|| ( player_info -> recompute_end ))
				{
					int		rc ;

					pos = player_info -> read_position ;
					rc = ifo_data_follower ( player_info , & pos ,
								& until , & player_info -> read_title , 1 , 
								player_info -> recompute_end ? 
									IFO_DFMODE_FIND_END : IFO_DFMODE_FIND_NEXT ) ;
					if ( rc == IFO_OK )
					{
						if ( ! player_info -> recompute_end )
						{
							if ( pos != player_info -> read_position )
							{
								pthread_mutex_lock ( & ( player_info->file_access_mutex ) ) ;
								pbc_seek ( player_info , pos ) ;
								pthread_mutex_unlock ( & ( player_info->file_access_mutex ) ) ;
								player_info -> read_position = pos ;
								player_info -> read_position_sub = 0 ;
							}
						} else {
							player_info -> recompute_end = 0 ;
						}
					} else {
						char		text [ 32 ] ;
						
						switch ( rc )
						{
							case IFO_NOT_AVAILABLE :
							case IFO_NO_TABLE :
							case IFO_SILLY_DATA :
									sprintf ( text , "%d" , rc ) ;
									msg_send_message ( player_info ,
											MSG_IFO_ERROR , text ) ;
									/* stop IFO parsing. */
									player_info -> read_title = -1 ;
									break ;
						}
					}
					last_pos = pos ;
				}
			}

			/* tell write thread that the buffer is full */
			if (( player_info -> writer_waiting )
			&& ( queue_quite_full ( player_info -> buffer ) ))
			{
				pthread_cond_broadcast ( & ( player_info -> buffer_quite_full ) ) ;
			}
		}

		pthread_cond_wait ( & ( player_info -> reader_state_change ) ,
				& ( player_info -> mutex ) ) ;
	}
	pthread_mutex_unlock ( & ( player_info->mutex ) ) ;
	return NULL ;
}
		

void * buf_writer ( void * arg )
{
	player_info_t 	* player_info = (player_info_t *) arg ;
	buffer_item		* buffer_chunk ;
	int				lba_authed = 0 ;
	int				current_type = TYPE_VOB ;
	int				fra = 0 ;
	char			err [ BUFFER_LENGTH ] ;

	pthread_mutex_lock ( & ( player_info->mutex ) ) ;
	while ( ! player_info -> b_quit )
	{
		if ( player_info -> b_running )
		{
			if (( buffer_chunk = (buffer_item *)
					queue_remove_from_queue ( player_info -> buffer )) != NULL )
			{
				if ( player_info -> force_reauth )
				{
					fra = 1 ;
					player_info -> force_reauth = 0 ;
				}
				pthread_mutex_unlock ( & ( player_info -> mutex ) ) ;

				/* Change the output A/V settings to match current
				 * stream type if necessary */
				if ( buffer_chunk -> source_type != current_type )
				{
					/* change audio/video settings to match type. */
					if ( output_get_playmode ( player_info -> out_plugin ) == OUTPUT_PLAYMODE_PLAY )
					{
						output_set_playmode ( player_info -> out_plugin , OUTPUT_PLAYMODE_STOPPED ) ;

						pthread_mutex_lock ( & ( player_info -> audio_mutex ) ) ;
						change_type_settings ( player_info -> out_plugin , 
								buffer_chunk -> source_type ,
								player_info -> default_audio_format ,
								player_info -> default_audio_width ,
								player_info -> default_audio_freq ,
								player_info -> default_video_format ) ;
						pthread_mutex_unlock ( & ( player_info -> audio_mutex ) ) ;

						output_set_playmode ( player_info -> out_plugin , OUTPUT_PLAYMODE_PLAY ) ;
						current_type = buffer_chunk -> source_type ;
					}
				}

				/* (Re)do the CSS title key if necessary */
				if (( buffer_chunk -> source_lba != lba_authed )
				|| ( fra ))
				{
					if ( buffer_chunk -> source_lba == 0 )
					{
						output_decryption_switch ( player_info -> out_plugin , 0 ) ;
					} else {
						if ( output_get_playmode ( player_info -> out_plugin ) == OUTPUT_PLAYMODE_PLAY )
						{
							output_set_playmode ( player_info -> out_plugin , OUTPUT_PLAYMODE_STOPPED ) ;

							if ( output_css_do_title_key 
								( player_info -> out_plugin , 
								err , BUFFER_LENGTH_1 , 
								buffer_chunk -> source_lba ) == OUTPUT_CSS_ERROR )
							{
								msg_send_message ( player_info ,
										MSG_CSS_ERROR , err ) ;
							}

							output_set_playmode ( player_info -> out_plugin , OUTPUT_PLAYMODE_PLAY ) ;
						}
					}
					lba_authed = buffer_chunk -> source_lba ;
					fra = 0 ;
				}

				/* this will block if the stream is bad, which, umm sucks */
				if ( output_write ( player_info -> out_plugin , buffer_chunk -> data , buffer_chunk -> size ) 
						< buffer_chunk -> size )
				{
					if ( errno != EINTR )
					{
						msg_send_message ( player_info ,
								MSG_WRITE_ERROR , strerror ( errno ) ) ;
					}
				}

				pthread_cond_broadcast ( & ( player_info -> reader_state_change ) ) ;

				pthread_mutex_lock ( & ( player_info -> mutex ) ) ;
				/* source is set initially by the read function */
				player_info -> write_position = buffer_chunk -> source_pos ;
				/* reattach buffer_chunk to standby buffer */
				assert ( queue_add_to_queue 
							( player_info -> standby_buffer , buffer_chunk ) ) ;
			} else {
				player_info -> writer_waiting = 1 ;
				pthread_cond_wait ( & ( player_info -> buffer_quite_full ) ,
							& ( player_info -> mutex ) ) ;
				player_info -> writer_waiting = 0 ;
			}
		} else {
			pthread_cond_wait ( & ( player_info -> reader_state_change ) ,
					& ( player_info -> mutex ) ) ;
		}
	}
	pthread_mutex_unlock ( & ( player_info->mutex ) ) ;
	return NULL ;
}

void buf_setup ( player_info_t * player_info )
{
	int			i ;

	pthread_mutex_init(&(player_info->mutex), NULL);
	pthread_mutex_init(&(player_info->audio_mutex), NULL);
	pthread_mutex_init(&(player_info->file_access_mutex), NULL);
	pthread_cond_init(&(player_info->reader_state_change), NULL);
	pthread_cond_init(&(player_info->buffer_quite_full), NULL);

	player_info -> num_files = 0 ;
	player_info -> read_fd = -1 ;
	player_info -> buffer = queue_make_queue ( BUFFER_RING_SIZE ) ;
	player_info -> standby_buffer = queue_make_queue ( BUFFER_RING_SIZE + 1 ) ;
	player_info -> b_running = 0 ;
	player_info -> b_quit = 0 ;
	player_info -> writer_waiting = 0 ;
	player_info -> read_title = -1 ;

	/* the buffer memory is preallocated. It sits in the standby buffer.
	 * When buffer memory is required the pointers are moved from the
	 * standby buffer into the main buffer, thus saving a calloc() every time */
	for ( i = 0 ; i < ( BUFFER_RING_SIZE + 1 ) ; i ++ )
	{
		assert ( queue_add_to_queue ( player_info -> standby_buffer ,
						calloc ( 1 , sizeof ( buffer_item ) ) ) ) ;
	}
}




static parser_table type_to_name [] = 
		{ { "mpg" , (void *) TYPE_MPEG , NULL , NONE } ,
		{ "mpeg" , (void *) TYPE_MPEG , NULL , NONE } ,
		{ "dat" , (void *) TYPE_MPEG , NULL , NONE } ,
		{ "vob" , (void *) TYPE_VOB , NULL , NONE } ,
		{ "m2v" , (void *) TYPE_VOB , NULL , NONE } ,
		{ "mpeg2" , (void *) TYPE_VOB , NULL , NONE } ,
		{ "mpg2" , (void *) TYPE_VOB , NULL , NONE } ,
		{ "" , NULL , NULL , 0 } } ;

int buf_get_type_from_name ( char * name )
{
	char		* ext ;
	int			type = TYPE_UNKNOWN ;

	ext = rindex ( name , '.' ) ;
	if ( ext == NULL )
	{
		return TYPE_UNKNOWN ;
	}

	if ( parse_type ( & ext [ 1 ] , type_to_name ,
			(void **) & type , NULL , NULL , NULL ) == NONE )
	{
		return type ;
	} else {
		return TYPE_UNKNOWN ;
	}
}


static void change_type_settings ( output_plugin * out , int type ,
			int def_format , int def_freq , int def_width ,
			int def_mode )
{
	if ( type == TYPE_MPEG )
	{
		output_set_video_type ( out , OUTPUT_TYPE_MPEG ) ;
		/* Use internal defaults for MPEG1 */
		output_select_audio ( out , OUTPUT_AUDIO_MPEG , 1 ) ;
		output_set_audio_freq ( out , OUTPUT_AUDIO_FREQ_441 ) ;
		output_set_audio_width ( out , OUTPUT_AUDIO_WIDTH_16 ) ;
		output_set_scaling_mode ( out , OUTPUT_ASPECTRATIOMODE_PAN_SCAN ) ;
	} else if ( type == TYPE_VOB )
	{
		output_set_video_type ( out , OUTPUT_TYPE_VOB ) ;
		output_set_audio_freq ( out , def_freq ) ;
		output_set_audio_width ( out , def_width ) ;
		output_set_scaling_mode ( out , def_mode ) ;
		output_select_audio ( out , def_format , 1 ) ;
	}
}


int buf_is_on_dvd ( player_info_t * player_info , char * filename )
{
	char		* dvd_video_ts_dir ;
	char		video_ts_part[] = "/video_ts/vts_" ;
	int			rc ;

	dvd_video_ts_dir = malloc ( strlen ( player_info -> dvd_mountpoint )
					+ strlen ( video_ts_part ) + 3 ) ;
	strcpy ( dvd_video_ts_dir , player_info -> dvd_mountpoint ) ;
	remove_trailing_slashes ( dvd_video_ts_dir ) ;
	strcat ( dvd_video_ts_dir , video_ts_part ) ;
	
	rc = ( strncasecmp ( dvd_video_ts_dir , filename ,
							strlen ( dvd_video_ts_dir ) ) == 0 ) ;

	free ( dvd_video_ts_dir ) ;

	return rc ;
}


