/*
 * Minidxr2 - Linux DVD player. Mostly (C) Jack Whitham, 2001
 *
 */

/*
 * control.c
 *
 * Implements a series of functions for DVD control from a thread. The functions
 * include a state machine implementation intended to prevent the caller from
 * 'breaking' the dvd player by doing something unexpected.
 *
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/stat.h>
#include <assert.h>

#include "control.h"
#include "parser.h"
#include "buffer.h"
#include "device.h"
#include "player.h"
#include "ifo_parse.h"
#include "read_config.h"
#include "message.h"

static char * handle_ifo_error ( int rc )
{
	switch ( rc )
	{
		case IFO_NOT_AVAILABLE :
				return "This feature is only available when a DVD title is loaded.\n" ;
		case IFO_SILLY_PARAMETERS :
				return "Invalid parameters given.\n" ;
		case IFO_SILLY_DATA :
		case IFO_NO_TABLE :
				return "Unable to read timecode table for this title.\n" ;
		default :
				return "IFO parsing error.\n" ;
	}
}

int dvd_open_file ( player_info_t * player_info , char ** file_list , int num_files ,
			char * error , int max_chars )
{
	char		* err ;

	if ( error == NULL )
	{
		return 0 ;
	}

	err = dvd_unload ( player_info ) ;
	if ( err != NULL )
	{
		SAFE_STRNCPY ( error , err , max_chars ) ;
		return 0 ;
	}

	err = buf_open_files ( player_info , file_list , num_files ) ;
	if ( err != NULL )
	{
		player_info -> h_state = DVD_UNLOADED ;
		SAFE_STRNCPY ( error , err , max_chars ) ;
		return 0 ;
	} else {
		if ( buf_files_on_dvd ( player_info ) )
		{
			/* We may need to use some CSS */
			if ( output_css_authenticate 
						( player_info -> out_plugin , 
						  player_info -> dvd_device , error , max_chars ) 
						== OUTPUT_CSS_OK )
			{
				if ( buf_get_lbas ( player_info , error , max_chars ) )
				{
					if ( output_css_do_disc_key 
							( player_info -> out_plugin , error ,
							  max_chars ) != OUTPUT_CSS_ERROR )
					{
						player_info -> h_state = DVD_STOPPED ;

						/* Do IFO parsing to work out the best
						 * starting audio track here. */
						dvd_set_track ( player_info , 0 ) ;
						return 1 ;
					}
				}
			}
			player_info -> h_state = DVD_STOPPED ;
			err = dvd_unload ( player_info ) ;
			return 0 ;
		} else {
			player_info -> h_state = DVD_STOPPED ;
			return 1 ;
		}
	}
}

char * dvd_unload ( player_info_t * player_info )
{
	char		* err ;
	int			on_dvd ;

	if ( player_info -> h_state == DVD_UNLOADED )
	{
		return NULL ;
	}

	if ( ifo_available ( player_info ) )
	{
		ifo_close ( player_info ) ;
	}

	on_dvd = buf_files_on_dvd ( player_info ) ;

	if ( player_info -> h_state != DVD_STOPPED )
	{
		err = dvd_stop ( player_info ) ;
		if ( err != NULL )
		{
			return err ;
		}
	}

	err = buf_close_files ( player_info ) ;

	if ( on_dvd )
	{
		output_css_finish ( player_info -> out_plugin ) ;
	}

	player_info -> h_state = DVD_UNLOADED ;

	return err ;
}

char * dvd_quit ( player_info_t * player_info )
{
	char		* err ;

	err = dvd_unload ( player_info ) ;

	(void) buf_playback_controller ( player_info , PBC_QUIT , NULL ) ;

	player_info -> h_state = DVD_QUIT ;

	return err ;
}

int dvd_open_title ( player_info_t * player_info , int t ,
			char * error , int max_chars )
{
	struct stat statbuf ;
	int			i , j ;
	char		* fname [ 10 ] ;
	char		* err ;
	int			ret ;
	int			title , angle ;		/* do not use angle for now */

	if ( error == NULL )
	{
		return 0 ;
	}

	err = dvd_unload ( player_info ) ;
	if ( err != NULL )
	{
		SAFE_STRNCPY ( error , err , max_chars ) ;
		return 0 ;
	}

	if ( ! ifo_open ( player_info ) )
	{
		SAFE_STRNCPY ( error , "Unable to read any IFO files.\n" , max_chars ) ;
		return 0 ;
	}

	title = t ;
	angle = 0 ;

	if (( title < 1 )
	|| ( title > 99 ))
	{
		/* temporarily open the IFO files in order to
		 * determine a place to start */

		ret = ifo_get_pos_from_main_title_search_pointer ( 
				player_info , 0 , & title , 
				& angle ) ;

		if (( ret != IFO_OK )
		|| ( title > 99 )
		|| ( title < 1 ))
		{
			ifo_close ( player_info ) ;
			SAFE_STRNCPY ( error , "Unable to read the disc IFO file.\n" , max_chars ) ;
			return 0 ;
		}
	}

	for ( i = 1 ; i < 10 ; i ++ )
	{
		char		buf1 [ 32 ] ;
		char		buf2 [ FNAME_LENGTH ] ;

		snprintf ( buf1 , 31 , TITLE_FILENAME , title , i ) ;
		buf1 [ 31 ] = '\0' ;
		strcpy ( buf2 , player_info -> dvd_mountpoint ) ;
		strcat ( buf2 , buf1 ) ;
		if ( stat ( buf2 , & statbuf ) != 0 )
		{
			string_uppercase ( buf1 ) ;
			strcpy ( buf2 , player_info -> dvd_mountpoint ) ;
			if (( strlen ( buf1 ) + strlen ( buf2 ) ) >= FNAME_LENGTH_1 )
			{
				ifo_close ( player_info ) ;
				SAFE_STRNCPY ( error , "The filenames are too long.\n" , max_chars ) ;
				return 0 ;
			}
			strcat ( buf2 , buf1 ) ;
			if ( stat ( buf2 , & statbuf ) != 0 )
			{
				break ;
			} else {
				fname [ i - 1 ] = strdup ( buf2 ) ;
			}
		} else {
			fname [ i - 1 ] = strdup ( buf2 ) ;
		}
	}

	if ( i < 2 )
	{
		ifo_close ( player_info ) ;
		SAFE_STRNCPY ( error , "No files found in that title.\n" , max_chars ) ;
		return 0 ;
	}

	ret = dvd_open_file ( player_info , fname , i - 1 , error , max_chars ) ;

	for ( j = 1 ; j < i ; j ++ )
	{
		free ( fname [ j - 1 ] ) ;
	}

	if ( ret )
	{
		int pos ;

		player_info -> current_title = title ;
		buf_playback_controller ( player_info , PBC_SET_TITLE , & title ) ;

		/* start from chapter 1 (The Shining fix) */
		pos = ifo_get_chapter_position ( player_info ,
				player_info -> current_title , 1 ) ;

		buf_playback_controller ( player_info , PBC_SEEK , & pos ) ;
	} else {
		ifo_close ( player_info ) ;
		player_info -> current_title = 0 ;
	}

	return ret ;
}


char * dvd_ffwd ( player_info_t * player_info )
{
	if ( player_info -> h_state == DVD_PLAYING )
	{
		output_fast_forwards( player_info -> out_plugin , OUTPUT_PLAYRATE_3x ) ;
		player_info -> h_state = DVD_FAST_FORWARD ;
		return NULL ;
	} else if ( player_info -> h_state == DVD_FAST_FORWARD )
	{
		return dvd_play ( player_info ) ;
	} else {
		return "Can't fast forward (not playing)\n" ;
	}
}

char * dvd_get_chapter ( player_info_t * player_info , int * output )
{
	char	* err ;
	int		i , startpos , endpos , curpos ;
	int		num = ifo_get_num_chapters ( player_info , 
				player_info -> current_title ) ;

	if (( player_info -> h_state == DVD_UNLOADED )
	|| ( player_info -> h_state == DVD_QUIT ))
	{
		return "This information is only available when a DVD is loaded.\n" ;
	}
	if ( ! ifo_available ( player_info ) )
	{
		return "Chapter numbers can only be obtained when a DVD title is loaded.\n" ;
	}

	curpos = 0 ;
	err = dvd_get_position ( player_info , & curpos ) ;
	if ( err != NULL )
	{
		return err ;
	}

	if ( num > 0 )
	{
		startpos = 0 ;
		for ( i = 0 ; i < num ; i ++ )
		{
			if (( i + 1 ) < num )
			{
				endpos = ifo_get_chapter_position ( player_info ,
						player_info -> current_title , i + 2 ) ;
			} else {
				/* not last chapter */
				endpos = 0 ;
				err = dvd_get_length ( player_info , & endpos ) ;
				if ( err != NULL )
				{
					return err ;
				}
			}
			if (( startpos <= curpos )
			&& ( curpos < endpos ))
			{
				(* output) = i + 1 ;
				return NULL ;
			}

			startpos = endpos ;
		}
		return "The current chapter number is unknown\n" ;
	} else if ( num == 0 )
	{
		return "No chapters found." ;
	} else {
		return handle_ifo_error ( num ) ;
	}
}

char * dvd_get_length ( player_info_t * player_info , int * output )
{
	if (( player_info -> h_state != DVD_UNLOADED )
	&& ( player_info -> h_state != DVD_QUIT ))
	{
		buf_playback_controller ( player_info , PBC_GET_LENGTH , output ) ;
		return NULL ;
	} else {
		return "Can't get length (no file loaded)\n" ;
	}
}

char * dvd_get_position ( player_info_t * player_info , int * output )
{
	if (( player_info -> h_state != DVD_UNLOADED )
	&& ( player_info -> h_state != DVD_QUIT ))
	{
		buf_playback_controller ( player_info , PBC_GET_POS , output ) ;
		return NULL ;
	} else {
		return "Can't get length (no file loaded)\n" ;
	}
}

char * dvd_get_time_code ( player_info_t * player_info , char * output , int max_chars )
{
	int		rc ;
	char	* ret ;
	int		hour , min , sec , pos ;

	if ( ! ifo_available ( player_info ) )
	{
		return "Time codes can only be obtained when a DVD title is loaded.\n" ;
	}

	ret = dvd_get_position ( player_info , & pos ) ;
	if ( ret != NULL )
	{
		return ret ;
	}

	rc = ifo_get_time_for_position ( player_info ,
			player_info -> current_title , pos , & hour , & min , & sec ) ;


	if ( rc == IFO_OK )
	{
		snprintf ( output , max_chars - 1 , "%d:%02d:%02d\n" , hour , min , sec ) ;
		output [ max_chars - 1 ] = '\0' ;

		return NULL ;
	} else {
		return handle_ifo_error ( rc ) ;
	}
}

char * dvd_pause ( player_info_t * player_info )
{
	char		* err ;

	if (( player_info -> h_state == DVD_PLAYING )
	|| ( player_info -> h_state == DVD_FAST_FORWARD ))
	{
		output_set_playmode ( player_info -> out_plugin , OUTPUT_PLAYMODE_PAUSED ) ;
		err = buf_playback_controller ( player_info , PBC_STOP , NULL ) ;
		if ( err == NULL )
		{
			player_info -> h_state = DVD_PAUSED ;
		}
		return err ;
	} else if ( player_info -> h_state == DVD_PAUSED )
	{
		return dvd_play ( player_info ) ;
	} else {
		return "Can't pause (not playing)\n" ;
	}
}

char * dvd_play ( player_info_t * player_info )
{
	char		* err ;

	if (( player_info -> h_state == DVD_STOPPED )
	|| ( player_info -> h_state == DVD_PAUSED ))
	{
		output_set_playmode ( player_info -> out_plugin , OUTPUT_PLAYMODE_PLAY ) ;
		err = buf_playback_controller ( player_info , PBC_START , NULL ) ;
		if ( err == NULL )
		{
			player_info -> h_state = DVD_PLAYING ;
		}
		return err ;
	} else if ( player_info -> h_state == DVD_FAST_FORWARD )
	{
		output_set_playmode ( player_info -> out_plugin , OUTPUT_PLAYMODE_PLAY ) ;
		player_info -> h_state = DVD_PLAYING ;
		return NULL ;
	} else if ( player_info -> h_state == DVD_PLAYING )
	{
		return NULL ;
	} else {
		return "Can't play\n" ;
	}
}

char * dvd_print_total_time ( player_info_t * player_info , char * output , int max_chars )
{
	int		rc ;
	char	* ret ;
	int		hour , min , sec , pos ;

	if ( ! ifo_available ( player_info ) )
	{
		return "Total time can only be obtained when a DVD title is loaded.\n" ;
	}

	ret = dvd_get_length ( player_info , & pos ) ;
	if ( ret != NULL )
	{
		return ret ;
	}

	rc = ifo_get_time_for_position ( player_info ,
			player_info -> current_title , pos , & hour , & min , & sec ) ;

	if ( rc == IFO_OK )
	{
		snprintf ( output , max_chars - 1 , "%d:%02d:%02d\n" , hour , min , sec ) ;
		output [ max_chars - 1 ] = '\0' ;

		return NULL ;
	} else {
		return handle_ifo_error ( rc ) ;
	}
}

char * dvd_print_audio_info ( player_info_t * player_info , char * output , int max_chars )
{
	char		* langname = NULL ;
	char		* audio_format_str ;
	char		* audio_freq_str ;
	char		* audio_width_str ;
	int			langcode ;
	int			num_channels ;
	int			cs ;
	char		buf [ BUFFER_LENGTH ] ;

	if ( ! ifo_available ( player_info ) )
	{
		return "Audio info can only be obtained when a DVD title is loaded.\n" ;
	}

	output [ 0 ] = '\0' ;
	for ( cs = 0 ; cs < ifo_get_num_audio_streams ( player_info ,
								player_info -> current_title ) ; cs ++ )
	{
		if (( parse_reverse ( ifo_settings_audio_width ( player_info , 
					player_info -> current_title , cs ) ,
						audio_width_table , & audio_width_str , NULL , NULL ))
		&& ( parse_reverse ( ifo_settings_audio_freq ( player_info , 
					player_info -> current_title , cs ) ,
						audio_freq_table , & audio_freq_str , NULL , NULL ))
		&& ( parse_reverse ( ifo_settings_audio_format ( player_info , 
					player_info -> current_title , cs ) ,
						audio_stream_table , & audio_format_str , NULL , NULL )))
		{
			langcode = ifo_settings_langcode ( player_info , 
						player_info -> current_title , cs ) ;

			if ( langcode > 0 )
			{
				langname = ifoDecodeLang ( langcode ) ;
			} else {
				langname = NULL ;
			}
			if (( langname == NULL )
			|| ( strlen ( langname ) == 0 ))
			{
				langname = "Unknown" ;
			}

			num_channels = ifo_settings_num_channels ( player_info ,
						player_info -> current_title , cs ) ;

			if ( num_channels > 0 )
			{
				snprintf ( buf , BUFFER_LENGTH_1 ,
						"track %d: %s ( %d channels %s bits %s kHz ) %s\n" ,
							cs + 1 , audio_format_str , num_channels , 
							audio_width_str , audio_freq_str , langname ) ;
				buf [ BUFFER_LENGTH_1 ] = '\0' ;

				if (( strlen ( output ) + strlen ( buf ) + 1 ) < max_chars )
				{
					strcat ( output , buf ) ;
				}
			} else {
				return handle_ifo_error ( num_channels ) ;
			}
		} else {
			return handle_ifo_error ( IFO_UNKNOWN ) ;
		}
	}
	return NULL ;
}

char * dvd_print_chapter_info ( player_info_t * player_info , char * output , int max_chars )
{
	int		i , hour , min , sec , rc ;
	int		pos ;
	char	* err ;
	char	buf [ BUFFER_LENGTH ] ;
	int		num = ifo_get_num_chapters ( player_info , 
				player_info -> current_title ) ;

	if (( player_info -> h_state == DVD_UNLOADED )
	|| ( player_info -> h_state == DVD_QUIT ))
	{
		return "This information is only available when a DVD is loaded.\n" ;
	}
	if ( ! ifo_available ( player_info ) )
	{
		return "Chapter numbers can only be obtained when a DVD title is loaded.\n" ;
	}


	if ( num > 0 )
	{
		output [ 0 ] = '\0' ;
		for ( i = 0 ; i < ( num + 1 ) ; i ++ )
		{
			if ( i < num )
			{
				/* not end */
				pos = ifo_get_chapter_position ( player_info ,
						player_info -> current_title , i + 1 ) ;
			} else {
				pos = 0 ;
				err = dvd_get_length ( player_info , & pos ) ;
				if ( err != NULL )
				{
					return err ;
				}
			}

			if ( pos >= 0 )
			{
				rc = ifo_get_time_for_position 
						( player_info , player_info -> current_title ,
						  pos , & hour , & min , & sec ) ;
				if ( rc != IFO_OK )
				{
					return handle_ifo_error ( rc ) ;
				} 
			} else {
				return handle_ifo_error ( pos ) ;
			}

			if ( i < num )
			{
				if ( i < 1 )
				{
					hour = min = sec = 0 ;
				}
				snprintf ( buf , BUFFER_LENGTH_1 ,
						"chapter %2d:   %2d:%02d:%02d\n" ,
							i + 1 , hour , min , sec ) ;
			} else {
				snprintf ( buf , BUFFER_LENGTH_1 ,
						"end of title: %2d:%02d:%02d\n" ,
							hour , min , sec ) ;
			}

			buf [ BUFFER_LENGTH_1 ] = '\0' ;
			if (( strlen ( output ) + strlen ( buf ) + 1 ) < max_chars )
			{
				strcat ( output , buf ) ;
			}
		}
		return NULL ;
	} else if ( num == 0 )
	{
		return "No chapters found." ;
	} else {
		return handle_ifo_error ( num ) ;
	}
}

char * dvd_print_sub_info ( player_info_t * player_info , char * output , int max_chars )
{
	char		* langname = NULL ;
	int			langcode ;
	int			cs ;
	int			type ;
	char		buf [ BUFFER_LENGTH ] ;
	char		* typename ;

	if ( ! ifo_available ( player_info ) )
	{
		return "Subtitle info can only be obtained when a DVD title is loaded.\n" ;
	}

	output [ 0 ] = '\0' ;
	for ( cs = 0 ; cs < ifo_get_num_sub_streams ( player_info ,
								player_info -> current_title ) ; cs ++ )
	{
		langcode = 0 ;
		type = ifo_settings_sub_langcode ( player_info , 
					player_info -> current_title , cs ,
					& langcode ) ;

		if ( type == IFO_NOT_AVAILABLE )
		{
			break ;
		}
		switch ( type )
		{
			case IFO_LANGCODE_COMMENTARY :	typename = "(commentary)" ;
											break ;
			case IFO_LANGCODE_UNKNOWN :		typename = "(unknown)" ;
											break ;
			default :						typename = "" ;
											break ;
		}

		if ( langcode > 0 )
		{
			langname = ifoDecodeLang ( langcode ) ;
		} else {
			break ;
		}
		if (( langname == NULL )
		|| ( strlen ( langname ) == 0 ))
		{
			langname = "Unknown" ;
		}

		snprintf ( buf , BUFFER_LENGTH_1 ,
				"track %d: %s %s\n" , cs + 1 , langname , typename ) ;
		buf [ BUFFER_LENGTH_1 ] = '\0' ;

		if (( strlen ( output ) + strlen ( buf ) + 1 ) < max_chars )
		{
			strcat ( output , buf ) ;
		}
	}
	return NULL ;
}

char * dvd_seek_chapter ( player_info_t * player_info , int param_int )
{
	int pos = ifo_get_chapter_position ( player_info ,
			player_info -> current_title , param_int ) ;

	if ( pos >= 0 )
	{
		return dvd_seek_position ( player_info , pos ) ;
	} else {
		return handle_ifo_error ( pos ) ;
	}
}

char * dvd_seek_position ( player_info_t * player_info , int param_int )
{
	char		* err ;

	if (( player_info -> h_state == DVD_PAUSED )
	|| ( player_info -> h_state == DVD_STOPPED ))
	{
		err = dvd_play ( player_info ) ;
		if ( err != NULL )
		{
			return err ;
		}
	}

	if (( player_info -> h_state == DVD_PLAYING )
	|| ( player_info -> h_state == DVD_STOPPED )
	|| ( player_info -> h_state == DVD_FAST_FORWARD ))
	{
		err = buf_playback_controller ( player_info , PBC_SEEK , & param_int ) ;
	} else {
		err = "Can't seek (no files loaded)\n" ;
	}
	return err ;
}

char * dvd_seek_time ( player_info_t * player_info , int hour , int min , int sec )
{
	int		pos ;

	if ( ! ifo_available ( player_info ) )
	{
		return "This is only possible when a DVD title is loaded.\n" ;
	}

	if (( player_info -> h_state == DVD_UNLOADED )
	|| ( player_info -> h_state == DVD_QUIT ))
	{
		return "Can't seek time (not loaded).\n" ;
	}

	pos = ifo_get_position_for_time ( player_info ,
			player_info -> current_title , hour , min , sec ) ;

	if ( pos < 0 )
	{
		return handle_ifo_error ( pos ) ;
	} else {
		return dvd_seek_position ( player_info , pos ) ;
	}
}

char * dvd_set_angle ( player_info_t * player_info , int angle )
{
	return "Not implemented\n" ;
}

char * dvd_set_sub ( player_info_t * player_info , int sub )
{
	if (( sub > 0 )
	&& ( sub < 33 ))
	{
		(void) output_subpicture ( player_info -> out_plugin , 1 ) ;
		(void) output_select_subpicture ( player_info -> out_plugin , sub - 1 ) ;

		if (( ifo_available ( player_info ) )
		&& ( player_info -> current_title > 0 ))
		{
			clut_t		* clut ;

			clut = (clut_t *) ifo_get_sub_clut ( player_info , player_info -> current_title ) ;
			if ( clut != NULL )
			{
				(void) output_set_subpicture_palette ( 
					player_info -> out_plugin , clut ) ;
			}
		}
	} else {
		(void) output_subpicture ( player_info -> out_plugin , 0 ) ;
	}
	return NULL ;
}

char * dvd_set_track ( player_info_t * player_info , int track )
{
	int			use_defaults = 0 ;
	int			width = 0 , freq = 0 , format = 0 ;

	if (( track < 1 )
	|| ( track > 8 ))
	{
		return "Invalid track number (must be 1 to 8)\n" ;
	} else {
		player_info -> current_stream = track ;

		/* here is where we should scan the IFO files for the
		 * format (if it is available) */
		if ( ifo_available ( player_info ) )
		{
			if ( track > ifo_get_num_audio_streams ( player_info ,
								player_info -> current_title ) )
			{
				return "Invalid track number (there aren't that many tracks on the DVD)\n" ;
			} else {
				if ( ! ((( width = ifo_settings_audio_width ( player_info , 
								player_info -> current_title , track - 1 )) > 0 )
					&& (( freq = ifo_settings_audio_freq ( player_info , 
								player_info -> current_title , track - 1 )) > 0 )
					&& (( format = ifo_settings_audio_format ( player_info , 
								player_info -> current_title , track - 1 )) > 0 )))
				{
					use_defaults = 1 ;
				}
			}
		} else {
			use_defaults = 1 ;
		}

		pthread_mutex_lock ( & ( player_info -> audio_mutex ) ) ;
		output_set_audio_width ( player_info -> out_plugin , use_defaults ? player_info -> default_audio_width : width ) ;
		output_set_audio_freq ( player_info -> out_plugin , use_defaults ? player_info -> default_audio_freq : freq ) ;
		output_select_audio ( player_info -> out_plugin , 
			use_defaults ? player_info -> default_audio_format : format , track ) ;
		pthread_mutex_unlock ( & ( player_info -> audio_mutex ) ) ;
		return NULL ;
	}
}

char * dvd_stop ( player_info_t * player_info )
{
	char		* err ;
	int			pos ;

	if (( player_info -> h_state == DVD_PLAYING )
	|| ( player_info -> h_state == DVD_PAUSED )
	|| ( player_info -> h_state == DVD_FAST_FORWARD ))
	{
		err = buf_playback_controller ( player_info , PBC_STOP , NULL ) ;
		if ( err == NULL )
		{
			player_info -> h_state = DVD_STOPPED ;
			output_set_playmode ( player_info -> out_plugin , OUTPUT_PLAYMODE_STOPPED ) ;
			output_reset ( player_info -> out_plugin ) ;
			/* does it need a reinit at this point? */

			pos = 0 ;
			(void) buf_playback_controller ( player_info , PBC_SEEK , & pos ) ;
		}
		return err ;
	} else {
		return "Can't stop (not playing)\n" ;
	}
}



char * dvd_set_default_audio_format ( player_info_t * player_info , int af )
{
	pthread_mutex_lock ( & ( player_info -> audio_mutex ) ) ;
	player_info -> default_audio_format = af ;
	pthread_mutex_unlock ( & ( player_info -> audio_mutex ) ) ;
	return NULL ;
}

char * dvd_set_default_audio_width ( player_info_t * player_info , int af )
{
	pthread_mutex_lock ( & ( player_info -> audio_mutex ) ) ;
	player_info -> default_audio_width = af ;
	output_set_audio_width ( player_info -> out_plugin , af ) ;
	pthread_mutex_unlock ( & ( player_info -> audio_mutex ) ) ;
	return NULL ;
}

char * dvd_set_default_audio_freq ( player_info_t * player_info , int af )
{
	pthread_mutex_lock ( & ( player_info -> audio_mutex ) ) ;
	player_info -> default_audio_freq = af ;
	output_set_audio_freq ( player_info -> out_plugin , af ) ;
	pthread_mutex_unlock ( & ( player_info -> audio_mutex ) ) ;
	return NULL ;
}

int dvd_get_message ( player_info_t * p , int * code , char ** data )
{
	int			rc , num ;

	rc = msg_get_message ( p , code , data ) ;
	if ( ! rc )
	{
		return 0 ;
	}

	switch ( (* code) )
	{
		case MSG_IFO_ERROR :
				/* decode IFO errors.
				 * They are sent through the message system
				 * as an integer in text format. */
				assert ( (* data) != NULL ) ;
				num = atoi ( (* data) ) ;
				free ( (* data) ) ;
				(* data) = strdup ( handle_ifo_error ( num ) ) ;
				break ;
	}
	return rc ;
}

char * dvd_list_languages ( void )
{
	return ifoListLanguages () ;
}

/* Languages. */

static int match_langcode ( int * langcodes , int num_langcodes , char * searching_for )
{
	int				i ;
	uint16_t		temp ;
	char			temp_str [ 3 ] ;

	temp_str [ 2 ] = '\0' ;
	for ( i = 0 ; i < num_langcodes ; i ++ )
	{
		temp = (uint16_t) langcodes [ i ] ;
		memcpy ( temp_str , & temp , 2 ) ;
		if ( strncmp ( temp_str , searching_for , 2 ) == 0 )
		{
			return i ;
		}
	}
	return -1 ;
}

static int do_matching ( int * langcodes , int num_langcodes , char * codes )
{
	int			rc ;
	char		* pos ;

	if ( num_langcodes < 1 )
	{
		return -1 ;
	}

	pos = codes ;
	while ( pos != NULL )
	{
		while (( strlen ( pos ) >= 2 )
		&& ( ! isalpha ( pos [ 0 ] ) ))
		{
			pos ++ ;
		}

		if ( strlen ( pos ) >= 2 )
		{
			if (( rc = match_langcode ( langcodes , num_langcodes , pos )) >= 0 )
			{
				return rc + 1 ;
			}
			pos = index ( pos , ' ' ) ;
		} else {
			pos = NULL ;
		}
	}
	return 1 ;
}

/*
 * dvd_find_best_sub_track and dvd_find_best_audio_track both take
 * a NULL terminated array of language codes in order of preference,
 * best first. The language codes are in the short format (e.g. 'en').
 *
 */

int dvd_find_best_sub_track ( player_info_t * player_info , char * codes )
{
	int			cs , lc , type ;
	int			langcode [ 32 ] ;

	if ( ! ifo_available ( player_info ) )
	{
		return -1 ;
	}

	for ( cs = 0 ; ( cs < ifo_get_num_sub_streams ( player_info ,
								player_info -> current_title ))
					&& ( cs < 32 ) ; cs ++ )
	{
		lc = -1 ;
		type = ifo_settings_sub_langcode ( player_info , 
					player_info -> current_title , cs ,
					& lc ) ;
		if ( type == IFO_NOT_AVAILABLE )
		{
			break ;
		}
		langcode [ cs ] = lc ;
	}

	return do_matching ( langcode , cs , codes ) ;
}

int dvd_find_best_audio_track ( player_info_t * player_info , char * codes )
{
	int			cs ;
	int			langcode [ 8 ] ;

	if ( ! ifo_available ( player_info ) )
	{
		return -1 ;
	}

	for ( cs = 0 ; ( cs < ifo_get_num_audio_streams ( player_info ,
								player_info -> current_title ))
					&& ( cs < 8 ) ; cs ++ )
	{
		langcode [ cs ] = ifo_settings_langcode ( player_info , 
					player_info -> current_title , cs ) ;
	}

	return do_matching ( langcode , cs , codes ) ;
}


char * dvd_set_view_mode ( player_info_t * player_info , int mode )
{
	geom_t		geom ;

	if ( mode != player_info -> view_mode )
	{
		memset ( & geom , 0 , sizeof ( geom_t ) ) ;
		if ( output_set_view_mode ( player_info -> out_plugin , mode , geom ))
		{
			player_info -> view_mode = mode ;
			return NULL ;
		} else {
			return "Mode change failed.\n" ;
		}
	} else {
		return NULL ;
	}
}

char * dvd_initialise_overlay ( player_info_t * player_info , output_extra_info * ck )
{
	if ( player_info -> view_mode_changeable )
	{
		return NULL ; /* already initialised */
	}
	if (( player_info -> h_state == DVD_UNLOADED )
	|| ( player_info -> h_state == DVD_STOPPED ))
	{
		dvd_set_view_mode ( player_info , OUTPUT_VIEW_OFF ) ;
		if ( output_initialise_overlay ( player_info -> out_plugin , ck ) )
		{
			player_info -> view_mode_changeable = 1 ;
			return NULL ;
		} else {
			player_info -> view_mode_changeable = 0 ;
			return NULL ;	/* Not an error, really.. overlay not available. */
		}
	} else {
		return "Must be stopped in order to set up the overlay.\n" ;
	}
}


