/*
 * Minidxr2 - Linux DVD player. Mostly (C) Jack Whitham, 2001
 *
 */
/*
 * DVD device operations 
 *
 */

#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/stat.h>

#ifndef FIBMAP
#define FIBMAP 1
#endif


#include "player.h"
#include "control.h"
#include "device.h"

static char * find_dvdtitleauth ( void ) ;
static int device_get_lba_with_root_perms ( char * filename , int * lba ,
			char * error , int max_chars ) ;
static char * device_ioctl ( player_info_t * player_info , int number ) ;

/*
 * this function was modified from my_system in system(3)
 */
int device_system_call ( char * command , char * parameter , char * return_text , int return_text_len ) 
{
	int status;
	pid_t pid ;
	int retpipe [ 2 ] ;

	if ( return_text != NULL )
	{
		if ( pipe ( retpipe ) != 0 )
		{
			return -1 ;
		}
	}

	pid = fork();
	if ( pid < 0 )
	{
		return -1 ;
	} else if ( pid == 0 ) 
	{
		char * argv [ 3 ] ;
		char * environment [ 2 ] ;

		close ( STDIN_FILENO ) ;
		close ( STDOUT_FILENO ) ;
		close ( STDERR_FILENO ) ;

		if ( return_text != NULL )
		{
			close ( retpipe [ 0 ] ) ;
			dup2 ( retpipe [ 1 ] , STDOUT_FILENO ) ;
			dup2 ( retpipe [ 1 ] , STDERR_FILENO ) ;
		}

		argv [ 0 ] = command ;
		argv [ 1 ] = parameter ;
		argv [ 2 ] = NULL ;
		environment [ 0 ] = "PATH=/bin:/sbin" ;
		environment [ 1 ] = NULL ;
		execve( command , argv , environment ) ;
		exit ( 127 ) ;
	}

	if ( return_text != NULL )
	{
		close ( retpipe [ 1 ] ) ;
	}

	status = 0 ;
	while ( waitpid(pid, &status, 0) < 0 )
	{
		if (errno != EINTR)
		{
			if ( return_text != NULL )
			{
				close ( retpipe [ 0 ] ) ;
			}
			return -1;
		}
	}

	if ( return_text != NULL )
	{
		int		len ;

		len = read ( retpipe [ 0 ] , return_text , return_text_len - 1 ) ;
		return_text [ len >= 0 ? len : 0 ] = '\0' ;
		close ( retpipe [ 0 ] ) ;
	}
	return status ;
}

char * device_mount ( player_info_t * player_info )
{
	/* Close tray first. This is done automatically in Linux 2.2. */
	device_ioctl ( player_info , CDROMCLOSETRAY ) ;

	if ( strlen ( player_info -> dvd_mountpoint ) > 0 )
	{
		int		rc ;

		rc = device_system_call ( "/bin/mount" , player_info -> dvd_mountpoint , NULL , 0 ) ;
		if ( rc < 0 )
		{
			return "Unable to call mount program.\n" ;
		} else if ( rc > 0 )
		{
			return "Unable to mount DVD.\n" ;
		} else {
			return NULL ;
		}
	} else {
		return "Unable to mount DVD - DVD mountpoint not set in config file.\n" ;
	}
}

char * device_unmount ( player_info_t * player_info )
{
	if ( strlen ( player_info -> dvd_mountpoint ) > 0 )
	{
		int		rc ;

		rc = device_system_call ( "/bin/umount" , player_info -> dvd_mountpoint , NULL , 0 ) ;
		if ( rc < 0 )
		{
			return "Unable to call umount program.\n" ;
		} else if ( rc > 0 )
		{
			return "Unable to unmount DVD.\n" ;
		} else {
			return NULL ;
		}
	} else {
		return "Unable to unmount DVD - DVD mountpoint not set in config file.\n" ;
	}
}


static char * device_ioctl ( player_info_t * player_info , int number )
{
	if ( strlen ( player_info -> dvd_device ) > 0 )
	{
		int		fd ;

		/* The CSS side may also have opened the dvd device. But it may
		 * not have done (not always in use). So it is opened here, separately. */
		fd = open ( player_info -> dvd_device , O_RDONLY | O_NONBLOCK ) ;
		if ( fd >= 0 )
		{
			if ( ioctl ( fd , number , 0 ) == 0 )
			{
				close ( fd ) ;
				return NULL ;
			} else {
				close ( fd ) ;
				return "Unable to eject DVD.\n" ;
			}
		} else {
			return "Unable to open DVD device.\n" ;
		}
	} else {
		return "Unable to eject DVD - DVD device not set in config file.\n" ;
	}
}

char * device_eject ( player_info_t * player_info )
{
	return device_ioctl ( player_info , CDROMEJECT ) ;
}


int device_get_lba ( player_info_t * player_info , char * error , int max_chars , 
				char * filename , int * lba )
{
	char		text_buf [ BUFFER_LENGTH ] ;
	int			rc ;
	char		* valid ;

	if ( error != NULL )
	{
		error [ 0 ] = '\0' ;
	}
	if ( geteuid () == 0 )
	{
		return device_get_lba_with_root_perms ( filename , lba ,
						error , max_chars ) ;
	}

	if ( player_info -> dvdtitleauth_path == NULL )
	{
		if ( ( player_info -> dvdtitleauth_path =
			find_dvdtitleauth () ) == NULL )
		{
			if ( error != NULL )
			{
				SAFE_STRNCPY ( error , "Path to dvdtitleauth program is not known.\n" , max_chars ) ;
			}
			return 0 ;
		}
	}

	rc = device_system_call ( player_info -> dvdtitleauth_path , filename , 
				text_buf , BUFFER_LENGTH_1 ) ;
	if ( rc < 0 )
	{
		if ( error != NULL )
		{
			SAFE_STRNCPY ( error , "Unable to run dvdtitleauth program to get title LBA.\n" , max_chars ) ;
		}
		return 0 ;
	} else {
		if ( strlen ( text_buf ) > 0 )
		{
			if ( strncmp ( SUBPROG_ERROR_CODE , text_buf , strlen ( SUBPROG_ERROR_CODE ) ) == 0 )
			{
				if ( error != NULL )
				{
					SAFE_STRNCPY ( error , & text_buf [ strlen ( SUBPROG_ERROR_CODE ) ] , max_chars ) ;
				}
				return 0 ;
			} else {
				valid = NULL ; 
				(* lba) = (int) strtol ( text_buf , & valid , 10 ) ;
				if ( valid != text_buf )
				{
					return 1 ;
				} else {
					if ( error != NULL )
					{
						SAFE_STRNCPY ( error , "dvdtitleauth program returned an invalid LBA\n" , max_chars ) ;
					}
					return 0 ;
				}
			}
		} else {
			if ( error != NULL )
			{
				SAFE_STRNCPY ( error , "dvdtitleauth program did not return anything useful\n" , max_chars ) ;
			}
		}
	}
	return 0 ;
}

static char * find_dvdtitleauth_at ( char * path )
{
	char			dvdtitleauth[] = "dvdtitleauth" ;
	char			* buffer ;
	struct stat 	statbuf ;

	buffer = malloc ( strlen ( path ) + strlen ( dvdtitleauth ) + 3 ) ;

	strcpy ( buffer , path ) ;
	strcat ( buffer , "/" ) ;
	strcat ( buffer , dvdtitleauth ) ;

	if ( stat ( buffer , & statbuf ) == 0 )
	{
		/* check suid bit, owner, and permissions on dvdtitleauth. 
		 * must be owned by root, setuid, and executable. */

		if (( statbuf . st_uid == 0 )
		&& ( statbuf . st_mode | S_ISUID )
		&& ( statbuf . st_mode | S_IFREG )
		&& ( statbuf . st_mode | S_IXOTH )
		&& ( statbuf . st_mode | S_IXUSR ))
		{
			return buffer ;
		}
	}
	free ( buffer ) ;
	return NULL ;
}


static char * find_dvdtitleauth ( void )
{
	char 			* path ;
	char 			* last_sep ;
	char 			* next_sep ;
	char			* out ;
	int				i ;
	char			hw_path [ 4 ][ 20 ] = { "/usr/local/bin" ,
							"/usr/bin" , "/usr/local/sbin" , "" } ;

	path = strdup ( getenv ( "PATH" ) ) ;
	last_sep = path ;

	while (( next_sep = index ( last_sep , ':' )) != NULL )
	{
		next_sep [ 0 ] = '\0' ;
		if (( out = find_dvdtitleauth_at ( last_sep )) != NULL )
		{
			free ( path ) ;
			return out ;
		} else {
			last_sep = & next_sep [ 1 ] ;
		}
	}

	out = find_dvdtitleauth_at ( last_sep ) ;
	free ( path ) ;

	if ( out != NULL )
	{
		return out ;
	}
	
	/* Try a few hardwired paths */

	i = 0 ;
	while ( strlen ( hw_path [ i ] ) > 0 )
	{
		out = find_dvdtitleauth_at ( hw_path [ i ] ) ;
		if ( out != NULL )
		{
			return out ;
		}
		i ++ ;
	}
	return NULL ;
}

/*
 * The following function is only used if
 * minidxr2 is running as root, so that there
 * is no need to call the external dvdtitleauth program
 * (which normally has to be done to get the LBA without
 * root permissions)
 *
 */
static int device_get_lba_with_root_perms ( char * filename , int * lba ,
			char * error , int max_chars )
{
	int			fd ;

	if ( ( fd = open ( filename , O_RDONLY ) ) < 1 )
	{
		SAFE_STRNCPY ( error , "File could not be opened.\n" ,
					max_chars ) ;
		return 0 ;
	}

	(* lba) = 0 ;
	if ( ioctl ( fd , FIBMAP , lba ) != 0 )
	{
		close ( fd ) ;
		SAFE_STRNCPY ( error , "Could not get LBA\n" , max_chars ) ;
		return 0 ;
	}
	close ( fd ) ;
	return 1 ;
}

