/*
 * Minidxr2 - Linux DVD player. Mostly (C) Jack Whitham, 2001
 *
 */
#include <config.h>
#ifdef HAVE_DXR2

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <error.h>
#include <errno.h>
#include <unistd.h>
#include <linux/cdrom.h>
#include <sys/ioctl.h>
#include <dxr2ioctl.h>
#include <string.h>

#include "output.h"
#include "player.h"
#include "dxr2-api.h"
#include "css.h"

/* local functions */
static int dxr2_css_reauthorise ( css_manager * css , 
			dxr2_status_info_t * p , char * error , int max_chars ) ;
static int dxr2_css_invalidate ( css_manager * css ) ;

/*
 * gets a new AGID and does some handshaking
 *
 */
int dxr2_css_authenticate ( css_manager * css , 
			dxr2_status_info_t * p , char * dvd_device , 
			char * error , int max_chars )
{
	if ( css -> css_ready )
	{
		SAFE_STRNCPY ( error , "CSS: already set up.\n" , max_chars ) ;
		return OUTPUT_CSS_ERROR ;
	}

	if ( strlen ( dvd_device ) > 0 )
	{
		css -> drive_fd = open ( dvd_device , O_RDONLY | O_NONBLOCK ) ;
		if ( css -> drive_fd < 0 )
		{
			SAFE_STRNCPY ( error , "CSS: cannot open dvd device.\n" , max_chars ) ;
			return OUTPUT_CSS_ERROR ;
		}
	} else {
		SAFE_STRNCPY ( error , "CSS: no dvd device named. Check config file.\n" , max_chars ) ;
		return OUTPUT_CSS_ERROR ;
	}

	if ( dxr2_css_reauthorise ( css , p , error , max_chars ) == OUTPUT_CSS_OK )
	{
		return OUTPUT_CSS_OK ;
	} else {
		close ( css -> drive_fd ) ;
		return OUTPUT_CSS_ERROR ;
	}
}


static int dxr2_css_invalidate ( css_manager * css )
{
	css -> css_ready = 0 ;
	css -> auth . type = DVD_INVALIDATE_AGID;
	if (ioctl(css -> drive_fd, DVD_AUTH, & css -> auth)) 
	{
		return OUTPUT_CSS_ERROR ;
	} else {
		return OUTPUT_CSS_OK ;
	}
}

/*
 * css_finish: Invalidates the AGID
 */
int dxr2_css_finish (css_manager * css )
{
	if ( ! css -> css_ready )
	{
		return OUTPUT_CSS_ERROR ;
	}

	if ( dxr2_css_invalidate ( css ) != OUTPUT_CSS_OK )
	{
		close ( css -> drive_fd ) ;
		return OUTPUT_CSS_ERROR ;
	}  
	close ( css -> drive_fd ) ;
	return OUTPUT_CSS_OK ;
}

/*
 * css_reauthorise: gets a new AGID, invalidating the old one if necessary.
 * This should only be called if the drive fd is open.
 */

static int dxr2_css_reauthorise ( css_manager * css , 
			dxr2_status_info_t * p , char * error , int max_chars )
{
	dvd_authinfo		* auth ;

	auth = & css -> auth ;

	if ( css -> css_ready )
	{
		/* invalidate the AGID first */
		dxr2_css_invalidate ( css ) ;
	}

	css -> css_ready = 1 ;

	/* get a new AGID */
	memset(auth, 0, sizeof(*auth));
	auth->type = DVD_LU_SEND_AGID;
	auth->lsa.agid = 0;
	if (ioctl(css -> drive_fd, DVD_AUTH, auth)) 
	{
		SAFE_STRNCPY ( error , "DVD_AUTH: cannot get AGID\n" , max_chars ) ;
		return OUTPUT_CSS_ERROR ;
	}

	// get the challenge from the card & send it to the drive
	//
	// No, absolutely no output_ioctl calls here, it'll go bottom up.

	if ( dxr2_ioctl ( p , DXR2_IOC_GET_CHALLENGE_KEY, auth->hsc.chal)) 
	{
		SAFE_STRNCPY ( error , "DVD_AUTH: failed get challenge from card\n" , max_chars ) ;
		dxr2_css_finish ( css ) ;
		return OUTPUT_CSS_ERROR ;
	}
	auth->type = DVD_HOST_SEND_CHALLENGE;
	if (ioctl(css -> drive_fd, DVD_AUTH, auth)) 
	{
		SAFE_STRNCPY ( error , "DVD_AUTH: failed send challenge to drive\n" , max_chars ) ;
		dxr2_css_finish ( css ) ;
		return OUTPUT_CSS_ERROR ;
	}


	// get the drive's response & send it to the card
	if (ioctl(css -> drive_fd, DVD_AUTH, auth)) 
	{
		SAFE_STRNCPY ( error , "DVD_AUTH: failed get response from drive\n" , max_chars ) ;
		dxr2_css_finish ( css ) ;
		return OUTPUT_CSS_ERROR ;
	}

	if ( dxr2_ioctl ( p , DXR2_IOC_SEND_RESPONSE_KEY, auth->lsk.key)) 
	{
		SAFE_STRNCPY ( error , "DVD_AUTH: failed send response to card\n" , max_chars ) ;
		dxr2_css_finish ( css ) ;
		return OUTPUT_CSS_ERROR ;
	}


	// get the drive challenge key & send it to the card
	auth->type = DVD_LU_SEND_CHALLENGE;
	if (ioctl(css -> drive_fd, DVD_AUTH, auth)) 
	{
		SAFE_STRNCPY ( error , "DVD_AUTH: failed get challenge from drive\n" , max_chars ) ;
		dxr2_css_finish ( css ) ;
		return OUTPUT_CSS_ERROR ;
	}
	if ( dxr2_ioctl ( p , DXR2_IOC_SEND_CHALLENGE_KEY, auth->lsc.chal)) 
	{
		SAFE_STRNCPY ( error , "DVD_AUTH: failed send challenge to card\n" , max_chars ) ;
		dxr2_css_finish ( css ) ;
		return OUTPUT_CSS_ERROR ;
	}


	// get the card's response key & send it to the drive
	if ( dxr2_ioctl ( p , DXR2_IOC_GET_RESPONSE_KEY, auth->hsk.key)) 
	{
		SAFE_STRNCPY ( error , "DVD_AUTH: failed get response from card\n" , max_chars ) ;
		dxr2_css_finish ( css ) ;
		return OUTPUT_CSS_ERROR ;
	}
	auth->type = DVD_HOST_SEND_KEY2;
	if (ioctl(css -> drive_fd, DVD_AUTH, auth)) 
	{
		SAFE_STRNCPY ( error , "DVD_AUTH: failed send response to card\n" , max_chars ) ;
		dxr2_css_finish ( css ) ;
		return OUTPUT_CSS_ERROR ;
	}

	return OUTPUT_CSS_OK ;
}


/**
 *
 * Receives and transmits the encrypted disk key to the card
 *
 */

int dxr2_css_do_disc_key ( css_manager * css , dxr2_status_info_t * p , char * error , int max_chars )
{
  dvd_struct dvdStructBuf;  
  dvd_authinfo * auth ;

	if ( ! css -> css_ready )
	{
		SAFE_STRNCPY ( error , "css_do_disc_key called when not css_ready.\n" , max_chars ) ;
		return OUTPUT_CSS_ERROR ;
	}

	css -> css_not_encrypted = 0 ;
	auth = & ( css -> auth ) ;

	/* get the disk key from drive, attempt 1 */
	dvdStructBuf.disckey.agid = auth->lsa.agid;
	dvdStructBuf.type = DVD_STRUCT_DISCKEY;
	memset(dvdStructBuf.disckey.value, 0, 2048);
	if ( ioctl(css -> drive_fd, DVD_READ_STRUCT, &(dvdStructBuf)) ) 
	{
		if ( dxr2_css_reauthorise ( css , p , error , max_chars ) != OUTPUT_CSS_OK )
		{
			return OUTPUT_CSS_ERROR ;
		}
		/* attempt 2 */
		dvdStructBuf.disckey.agid = auth->lsa.agid;
		dvdStructBuf.type = DVD_STRUCT_DISCKEY;
		if ( ioctl(css -> drive_fd, DVD_READ_STRUCT, &(dvdStructBuf)) ) 
		{
			SAFE_STRNCPY ( error , "cannot get disc key from drive.\n" , max_chars ) ;
			/* so we assume the disc is not encrypted at this point */
			css -> css_not_encrypted = 1 ;

			return( OUTPUT_CSS_NOT_ENCRYPTED );
		}
	}

	// send it to the card
	if ( dxr2_ioctl ( p , DXR2_IOC_SEND_DISC_KEY, dvdStructBuf.disckey.value)) 
	{
		SAFE_STRNCPY ( error , "cannot send key to card\n" , max_chars ) ;
		return OUTPUT_CSS_ERROR ;
	}

	// OK
	return( OUTPUT_CSS_OK );
}



/**
 *
 * Receives and transmits the encrypted title key to the card
 *
 */

int dxr2_css_do_title_key ( css_manager * css , dxr2_status_info_t * p ,
			char * error , int max_chars , int lba )
{
	dxr2_titleKey_t titleKeyBuf;
	dvd_authinfo * auth ;

	if ( ! css -> css_ready )
	{
		SAFE_STRNCPY ( error , "css_do_title_key called when not css_ready.\n" , max_chars ) ;
		return OUTPUT_CSS_ERROR ;
	}
	if ( css -> css_not_encrypted )
	{
		return OUTPUT_CSS_NOT_ENCRYPTED ;
	}

	auth = & ( css -> auth ) ;

	/* get the title key from drive, attempt 1 */
	auth->lstk.type = DVD_LU_SEND_TITLE_KEY;
	auth->lstk.lba = lba;
	if (ioctl(css -> drive_fd, DVD_AUTH, auth)) 
	{
		if ( dxr2_css_reauthorise ( css , p , error , max_chars ) != OUTPUT_CSS_OK )
		{
			return OUTPUT_CSS_ERROR ;
		}
		/* attempt 2 */
		auth->lstk.type = DVD_LU_SEND_TITLE_KEY;
		auth->lstk.lba = lba;
		if (ioctl(css -> drive_fd, DVD_AUTH, auth)) 
		{
			SAFE_STRNCPY ( error , "couldn't get title key from drive\n" , max_chars ) ;
			return OUTPUT_CSS_ERROR ;
		}
	}

	// copy it into the titleKey structure 
	// (since we've got the mysterious extra byte to cope with)
	memcpy(titleKeyBuf.key, auth->lstk.title_key, 5);

	// CGMS values
	titleKeyBuf.cgmsFlags = ((auth->lstk.cpm << 7) | 
	(auth->lstk.cp_sec << 6) | 
	(auth->lstk.cgms << 4));

	// send it to the card
	if ( dxr2_ioctl ( p , DXR2_IOC_SEND_TITLE_KEY, & titleKeyBuf )) 
	{
		SAFE_STRNCPY ( error , "couldn't send title key to card\n" , max_chars ) ;
		return OUTPUT_CSS_ERROR ;
	}

	// turn CSS mode on 
	if ( ! dxr2_decryption_switch_ext ( p , 1 ) )
	{ 
		SAFE_STRNCPY ( error , "couldn't turn decryption on\n" , max_chars ) ;
		return OUTPUT_CSS_ERROR ; 
	} 

	return OUTPUT_CSS_OK ;
}



/*
 * DVD Reset function
 *
 * From libcss
 */

char * dxr2_css_reset ( char * device )
{
	dvd_authinfo ai;
	int i;
	int fd ;

	fd = open(device, O_RDONLY | O_NONBLOCK);

	if (fd < 0) {
		return "unable to open dvd drive\n" ;
	}

	for (i = 0; i < 4; i++) 
	{
		memset(&ai, 0, sizeof(ai));
		ai.type = DVD_INVALIDATE_AGID;
		ai.lsa.agid = i;
		ioctl(fd, DVD_AUTH, &ai);
	}

	close ( fd ) ;

	return NULL ;
}


#endif

