/*
 * Minidxr2 - Linux DVD player. Mostly (C) Jack Whitham, 2001
 *
 */
#include <config.h>
#ifdef HAVE_DXR2

/*
 * setup.c
 *
 * Code to initially set up the dxr2.
 *
 */

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>

#include "output.h"
#include "dxr2-api.h"
#include "player.h"
#include "parser.h"
#include "setup.h"
#include "read_config.h"
#include "x11.h"

/* local functions */
char * install_firmware ( player_info_t * player_info , dxr2_status_info_t* dxr2_info ) ;
void init_dxr2_info(dxr2_status_info_t* dxr2_info) ;

void * dxr2_setup ( player_info_t* player_info , char ** err )
{
	dxr2_status_info_t	* dxr2_info ;
	char				fuser_output [ SHORT_BUFFER_LENGTH ] ;

	dxr2_info = calloc ( 1 , sizeof ( dxr2_status_info_t ) ) ;

	dxr2_info -> dxr2_current_playmode = DXR2_PLAYMODE_STOPPED ; 

	init_dxr2_info ( dxr2_info ) ;

	output_if_valid_int_then_set ( dxr2_output_xlat , 
				& dxr2_info -> video_freq . arg1 ,
				player_info -> default_video_freq ) ;
	output_if_valid_int_then_set ( dxr2_output_xlat , 
				& dxr2_info -> tv_format . arg ,
				player_info -> default_tv_format ) ;
	output_if_valid_int_then_set ( dxr2_output_xlat , 
				& dxr2_info -> scaling_mode . arg ,
				player_info -> default_video_format ) ;
	output_if_valid_int_then_set ( dxr2_output_xlat , 
				& dxr2_info -> audio_width . arg ,
				player_info -> default_audio_width ) ;
	output_if_valid_int_then_set ( dxr2_output_xlat , 
				& dxr2_info -> audio_freq . arg ,
				player_info -> default_audio_freq ) ;
	output_if_valid_int_then_set ( dxr2_output_xlat , 
				& dxr2_info -> audio_stream . arg1 ,
				player_info -> default_audio_format ) ;

	/* fuser check */
	/* This is a simplistic way to detect if anyone is using the dxr2 device.
	 * We barf if someone is. Really, the driver should prevent two different
	 * players using it at the same time. */
	if ( device_system_call ( "/sbin/fuser" , player_info -> dxr2_device , 
							fuser_output , SHORT_BUFFER_LENGTH_1 ) == 0 )
	{
		/* may be someone using it. */
		char * colon ;

		colon = index ( fuser_output , ':' ) ;
		if ( colon != NULL )
		{
			if ( atoi ( & colon [ 1 ] ) > 0 )
			{
				/* PID found. */
				(* err) = "The Dxr2 device is in use.\n" ;
				free ( dxr2_info ) ;
				return NULL ;
			}
		}
	}

    if ( ( dxr2_info -> dxr2_fd = 
					open ( player_info->dxr2_device, O_WRONLY)) < 0) 
	{
		(* err) = "Unable to open dxr2 device - check config file.\n" ;
		free ( dxr2_info ) ;
		return NULL ;
    }  
	
	(* err) = install_firmware ( player_info , dxr2_info ) ;
	if ( (* err) != NULL )
	{
		close ( dxr2_info -> dxr2_fd ) ;
		free ( dxr2_info ) ;
		return NULL ;
	}

	// reset the player
	dxr2_init ( dxr2_info ) ;
	dxr2_set_params ( dxr2_info ) ;
	dxr2_set_up_audio ( dxr2_info ) ;

	return dxr2_info ;
}


void init_dxr2_info(dxr2_status_info_t* dxr2_info)
{
  // Default values
  dxr2_info->tv_format.arg= DXR2_OUTPUTFORMAT_NTSC;
  // Set this to an invalid value so that we set it later based on tv_format
  dxr2_info->video_freq.arg1=-1;
  dxr2_info->video_freq.arg2=-1;
  dxr2_info->video_freq.arg3=-1;
  dxr2_info->output_aspect_ratio.arg= DXR2_ASPECTRATIO_4_3;
  dxr2_info->source_aspect_ratio.arg=-1;
  dxr2_info->scaling_mode.arg=DXR2_ASPECTRATIOMODE_LETTERBOX;
  dxr2_info->bitstream_type.arg=DXR2_BITSTREAM_TYPE_MPEG_VOB;
  dxr2_info->macro_vision.arg=DXR2_MACROVISION_OFF;
  dxr2_info->pixel_mode.arg=DXR2_PIXEL_CCIR601;
  dxr2_info->interlaced_mode.arg=DXR2_INTERLACED_ON;
  dxr2_info->x75ire_mode.arg=DXR2_75IRE_OFF;
  dxr2_info->volume.arg=19;
  dxr2_info->mute.arg=DXR2_AUDIO_MUTE_OFF;
  dxr2_info->audio_width.arg=DXR2_AUDIO_WIDTH_16;
  dxr2_info->audio_freq.arg=-1;
  dxr2_info->iec_output_mode.arg = DXR2_IEC958_ENCODED; 
  dxr2_info->subpicture.arg=DXR2_SUBPICTURE_OFF;
  dxr2_info->subpicture_stream_id.arg1=DXR2_STREAM_SUBPICTURE;
  dxr2_info->subpicture_stream_id.arg2=0;
  dxr2_info->audio_stream.arg1=-1;
  dxr2_info->audio_stream.arg2=0;
  dxr2_info->video_stream.arg1=DXR2_STREAM_VIDEO;
  dxr2_info->video_stream.arg2=0;

  /* VGA OVERLAY INFO */
  dxr2_info->overlay_crop.arg1=0; // left
  dxr2_info->overlay_crop.arg2=0; // right
  dxr2_info->overlay_crop.arg3=20; // top
  dxr2_info->overlay_crop.arg4=20;  // bottom
  dxr2_info->overlay_pos.arg1=81; // origin X
  dxr2_info->overlay_pos.arg2=36; // origin Y
  dxr2_info->overlay_dim.arg1=800; // size horiz
  dxr2_info->overlay_dim.arg2=600; // size vert
  dxr2_info->overlay_mode.arg=DXR2_OVERLAY_DISABLED;
  dxr2_info->gain.arg1 = 63; // common gain (0-63)
  dxr2_info->gain.arg2 = 26; // red-gain (0-63)
  dxr2_info->gain.arg3 = 40; // green-gain (0-63)
  dxr2_info->gain.arg4 = 40; // blue-gain (0-63)
  dxr2_info->picture.arg1 = 10; // gamma
  dxr2_info->picture.arg2 = 100; // contrast
  dxr2_info->picture.arg3 = 100; // brightness
  dxr2_info->picture.arg4 = 0; // saturation INVALID
}

char * install_firmware(player_info_t* player_info, dxr2_status_info_t * dxr2_info )
{
	int uCodeFD;
	int uCodeSize;
	dxr2_uCode_t* uCode;

	if ((uCodeFD = open(player_info->uCode_file, O_RDONLY)) < 0) 
	{
		return "Couldn't open microcode file. Check the configuration file settings\n" ;
	}
	uCodeSize = lseek(uCodeFD, 0, SEEK_END);
	if ((uCode = malloc(uCodeSize + 40)) == NULL) 
	{
		close(uCodeFD);
		return "Couldn't allocate memory for microcode file.\n" ;
	}
	lseek(uCodeFD, 0, SEEK_SET);
	if (read(uCodeFD, uCode+4, uCodeSize) != uCodeSize) 
	{
		close(uCodeFD);
		free(uCode);
		return "Couldn't read microcode file.\n" ;
	}
	close(uCodeFD);
	uCode->uCodeLength = uCodeSize;

	// upload ucode
	if (dxr2_install_firmware( dxr2_info , uCode)) 
	{
		free(uCode);
		return "Microcode upload failed.\n" ;
	}
	free(uCode);
	return NULL ;
}


char * dxr2_deinitialise ( dxr2_status_info_t * p )
{
	/* Invalidate CSS AGID here.. (NOT) */

#ifdef HAVE_X_OVERLAY_SUPPORT
	x11_close_window ( p ) ;
#endif

	dxr2_set_playmode ( p , DXR2_PLAYMODE_STOPPED ) ;

	if ( p -> dxr2_fd >= 0 )
	{
		close ( p -> dxr2_fd ) ;
	}
	free ( p ) ;
	return NULL ;
}

#endif

