/*
 * Minidxr2 - Linux DVD player. Mostly (C) Jack Whitham, 2001
 *
 */
#include <config.h>

#ifdef HAVE_DXR2
#ifdef HAVE_X_OVERLAY_SUPPORT

/*
 * x11.c
 *
 * contains various functions to support the X-based DVD overlay system
 *
 */

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xmu/Xmu.h>
#include <X11/Xlib.h>
#include <X11/extensions/xf86vmode.h>
#include "interface.h"

#include "player.h"
#include "dxr2-api.h"

#define MIN_RATIO			10

#define DEBUG_REPORT(name,rc)	\
				printf ( "%s: %s\n" , ( name ) , \
						(( rc ) == 0 ) ? "ok" : "failed!" ) ;

#define AllPointerEventMask \
	(ButtonPressMask | ButtonReleaseMask | \
	EnterWindowMask | LeaveWindowMask | \
	PointerMotionMask | PointerMotionHintMask | \
	Button1MotionMask | Button2MotionMask | \
	Button3MotionMask | Button4MotionMask | \
	Button5MotionMask | ButtonMotionMask | \
	KeymapStateMask)

typedef struct whitescreen_return_struct {
						int quit ;
						int error_occured ;
						int ready ;
						int timeout ;
						pthread_mutex_t mutex ;
						pthread_cond_t change ;
					} whitescreen_return ;
	
typedef struct timer_data_struct {
						XtAppContext		app ;
						XtIntervalId		timer_id ;
					} timer_data ;


/*
 * Overlay setup functions 
 */

static void * x11_whitescreen ( void * p ) ;

static int x11_get_screen_geometry ( int * width , int * height )
{
	Display* dsp = NULL;
	/* Screen* screen;
	Window rootWindow; */
	XF86VidModeModeLine modeLine;
	int dotClk;
	int	rc = 0 ;

	(* width) = (* height) = 0 ;

	/* open display */
	if (!(dsp = XOpenDisplay(NULL))) 
	{
		return 0 ;
	}

	/* get the screen */
	/* screen = DefaultScreenOfDisplay(dsp); */

	/* get the root window of it */
	/* rootWindow = RootWindowOfScreen(screen); */

	/* get the screen width & height */
	if ( XF86VidModeGetModeLine(dsp, DefaultScreen(dsp), &dotClk, &modeLine) )
	{
		(* width) = modeLine . hdisplay ;
		(* height) = modeLine . vdisplay ;
		rc = 1 ;
	}

	/* close display again */
	XCloseDisplay(dsp);

	return ( rc && ( (* width) > 0 ) && ( (* height) > 0 ) ) ;
}

int x11_calculate_overlay_params ( dxr2_status_info_t * dxr2_info ,
			int read_cache )
{
	int xRes;
	int yRes;
	dxr2_sixArg_t buf6;
	dxr2_oneArg_t buf1;
	pthread_t 			whitescreen_thread ;
	pthread_attr_t		thread_attrib ;
	whitescreen_return	wr ;
	int					stage1failure ;

	dxr2_info -> overlay_avail = 0 ;

	/* get screen res */
	if ( ! x11_get_screen_geometry (&xRes, &yRes) )
	{
	  return 0 ;
	}

	if (( read_cache )
	&& ( cache_available ( & dxr2_info -> vgaBuf ) )
	&& ( dxr2_info -> vgaBuf . xScreen == xRes )
	&& ( dxr2_info -> vgaBuf . yScreen == yRes )
	&& ( dxr2_info -> vgaBuf . ratio >= MIN_RATIO ) )
	{
		if ( dxr2_setup_overlay( dxr2_info , &dxr2_info->vgaBuf ) >= 0 )
		{
			dxr2_info -> overlay_avail = 1 ;
			return 1 ;
		}
	}

	memset ( & dxr2_info -> vgaBuf , 0 , sizeof ( dxr2_vgaParams_t ) ) ;
	buf6.arg1 = 0;
	buf6.arg2 = 0; 
	buf6.arg3 = 55;
	buf6.arg4 = 300;
	dxr2_ioctl ( dxr2_info , DXR2_IOC_SET_OVERLAY_CROPPING, &buf6);                    


	/* OK, start another thread to do the whitescreen thang. */
	wr . timeout = wr . quit = wr . ready = wr . error_occured = 0 ;
	pthread_mutex_init ( & ( wr . mutex ) , NULL ) ;
	pthread_cond_init ( & ( wr . change ) , NULL ) ;
	pthread_attr_init(&thread_attrib);
    pthread_attr_setschedpolicy(&thread_attrib, SCHED_OTHER);
	pthread_create ( & whitescreen_thread , & thread_attrib , 
			x11_whitescreen , (void *) ( & wr ) ) ;
	pthread_attr_destroy (&thread_attrib);

	/* wait for whitescreen to either return an error or a ready */
	pthread_mutex_lock ( & ( wr . mutex ) ) ;
	pthread_cond_wait ( & ( wr . change ) , & ( wr . mutex ) ) ;
	if ( wr . error_occured )
	{
		wr . quit = 1 ;
		pthread_mutex_unlock ( & ( wr . mutex ) ) ;
		pthread_join ( whitescreen_thread , NULL ) ;
		pthread_mutex_destroy ( & ( wr . mutex ) ) ;
		pthread_cond_destroy ( & ( wr . change ) ) ;
		return 0 ;
	}
	pthread_mutex_unlock ( & ( wr . mutex ) ) ;

	stage1failure = 0 ;
	/* set colour key */
	buf6.arg1 = 0x40; // Red min
	buf6.arg2 = 0xff; // Red Max
	buf6.arg3 = 0x40; // Green
	buf6.arg4 = 0xff;
	buf6.arg5 = 0x40; // Blue
	buf6.arg6 = 0xff;
	stage1failure |= 
		dxr2_ioctl ( dxr2_info , DXR2_IOC_SET_OVERLAY_COLOUR, &buf6) ;

	// setup window key
	buf6.arg1 = 1000;
	stage1failure |= 
		dxr2_ioctl ( dxr2_info , DXR2_IOC_SET_OVERLAY_RATIO, &buf6) ;

	buf6.arg1 = 100;
	buf6.arg2 = 3;
	stage1failure |= 
		dxr2_ioctl ( dxr2_info , DXR2_IOC_SET_OVERLAY_POSITION, &buf6) ;

	buf6.arg1 = xRes;
	buf6.arg2 = yRes;
	stage1failure |= 
		dxr2_ioctl ( dxr2_info , DXR2_IOC_SET_OVERLAY_DIMENSION, &buf6 ) ;

	// back to window keying
	buf1.arg = 3;
	stage1failure |= 
		dxr2_ioctl ( dxr2_info , DXR2_IOC_SET_OVERLAY_IN_DELAY, &buf1) ;


	// turn window AND colour keying on
	buf1.arg = DXR2_OVERLAY_WINDOW_COLOUR_KEY;
	stage1failure |= 
		dxr2_ioctl ( dxr2_info , DXR2_IOC_SET_OVERLAY_MODE, &buf1) ;

	DEBUG_REPORT ( "stage 1 setup" , stage1failure ) ;

	// calculate VGA parameters
	dxr2_info->vgaBuf.xScreen = xRes;
	dxr2_info->vgaBuf.yScreen = yRes ;
	dxr2_info->vgaBuf.hOffWinKey = 100;
	dxr2_info->vgaBuf.vOffWinKey = 3;

	/* these ioctls calibrate the rest of the vga parameters.. */
	DEBUG_REPORT ( "calc" , 
			dxr2_ioctl ( dxr2_info , 
				DXR2_IOC_CALCULATE_VGA_PARAMETERS, &dxr2_info->vgaBuf ) ) ;
	DEBUG_REPORT ( "set" , 
			dxr2_ioctl ( dxr2_info , 
				DXR2_IOC_SET_VGA_PARAMETERS , &dxr2_info->vgaBuf ) ) ;

	printf ( "The parameters are as follows:\n"
			"dxr2_info -> vgaBuf . hsyncPol = %d\n"
			"dxr2_info -> vgaBuf . blankStart = %d\n"
			"dxr2_info -> vgaBuf . blankWidth = %d\n"
			"dxr2_info -> vgaBuf . hOffset = %d\n"
			"dxr2_info -> vgaBuf . vOffset = %d\n"
			"dxr2_info -> vgaBuf . ratio = %d\n" , dxr2_info -> vgaBuf . hsyncPol ,
					dxr2_info -> vgaBuf . blankStart , dxr2_info -> vgaBuf . blankWidth, 
					dxr2_info -> vgaBuf . hOffset , dxr2_info -> vgaBuf . vOffset ,
					dxr2_info -> vgaBuf . ratio ) ;

	/* Stop the whitescreen thread */
	pthread_mutex_lock ( & ( wr . mutex ) ) ;
	wr . quit = 1 ;
	pthread_mutex_unlock ( & ( wr . mutex ) ) ;
	pthread_join ( whitescreen_thread , NULL ) ;
	pthread_mutex_destroy ( & ( wr . mutex ) ) ;
	pthread_cond_destroy ( & ( wr . change ) ) ;

	if (( dxr2_info -> vgaBuf . ratio < MIN_RATIO )
	|| ( wr . timeout ))
	{
		printf ( "The parameters seem to be wrong, overlay initialisation\n"
				"has failed.\n" ) ;
		/* looks like something went wrong. */
		return 0 ;
	} else {
		cache_write ( & dxr2_info -> vgaBuf ) ;

		/* OK, setup the overlay properly */
		if ( dxr2_setup_overlay( dxr2_info , &dxr2_info->vgaBuf ) >= 0 )
		{
			dxr2_info -> overlay_avail = 1 ;
			return 1 ;
		} else {
			return 0 ;
		}
	}
}

int x11_init_overlay_settings ( dxr2_status_info_t * dxr2_info )
{
	int xRes;
	int yRes;

	if ( ! x11_get_screen_geometry (&xRes, &yRes) )
	{
		return 0 ;
	}
	dxr2_info->color_key.arg1 = dxr2_info->color_key.arg3 = dxr2_info->color_key.arg5 = 0x40 ;
	dxr2_info->color_key.arg2 = dxr2_info->color_key.arg4 = dxr2_info->color_key.arg6 = 0xff ;
    dxr2_info->overlay_crop.arg1 = dxr2_info->overlay_crop.arg2 = 0 ;
    dxr2_info->overlay_crop.arg3 = 20 ;
	dxr2_info->overlay_crop.arg4 = 310 ;		/* a good default for my machine */
    dxr2_info->overlay_dim.arg1 = xRes ;
	dxr2_info->overlay_dim.arg2 = yRes ;
    dxr2_info->overlay_pos.arg1 = 0 ;
	dxr2_info->overlay_pos.arg2 = 0 ;
	return 1 ;
}
    
static void x11_whitescreen_timer ( XtPointer p , XtIntervalId * interval )
{
	int		* countdown ;

	countdown = (int *) p ;
	(* countdown) -- ;
	XtAddTimeOut ( 100 , x11_whitescreen_timer , p ) ;
}

/* x11_whitescreen: Function from old version of dxr2
 *
 * This was partly rewritten to use threads rather than fork,
 * and also because of my pathological hatred of gotos. 
 */

static void * x11_whitescreen ( void * p )
{
	whitescreen_return * wr = (whitescreen_return *) p ;
	Display* dsp = NULL;
	Window rootWindow;
	Window blackWindow;
	Screen* screen;
	XSetWindowAttributes xswa;
	Cursor myCursor;
	Pixmap lockc;
	Pixmap lockm;
	char no_bits[] = {0};
	int countdown ;
	XColor nullcolor;

	pthread_mutex_lock ( & ( wr -> mutex ) ) ;
	wr -> timeout = wr -> error_occured = wr -> ready = 0 ;
	pthread_mutex_unlock ( & ( wr -> mutex ) ) ;

	// open display
	if (!(dsp = XOpenDisplay(NULL))) 
	{
		pthread_mutex_lock ( & ( wr -> mutex ) ) ;
		wr -> error_occured = 1 ;
		pthread_mutex_unlock ( & ( wr -> mutex ) ) ;
		pthread_cond_broadcast ( & ( wr -> change ) ) ;
		return NULL ;
	}

	// get the screen
	screen = DefaultScreenOfDisplay(dsp);

	// get the root window of it
	rootWindow = RootWindowOfScreen(screen);

	// setup window attributes
	xswa.background_pixel = WhitePixelOfScreen(screen); // 0x0010#10e0;
	xswa.override_redirect = True;

	/* create blank cursor */
	lockc = XCreateBitmapFromData(dsp, rootWindow, no_bits, 1, 1);
	lockm = XCreateBitmapFromData(dsp, rootWindow, no_bits, 1, 1);
	myCursor = XCreatePixmapCursor(dsp, lockc, lockm, &nullcolor, &nullcolor, 0, 0);
	XFreePixmap(dsp, lockc);
	XFreePixmap(dsp, lockm);

	/* create & open the white fullscreen window */
	if ((blackWindow = XCreateWindow(dsp, rootWindow, 
									-2, -2, 
									WidthOfScreen(screen)+10, // just in case
									HeightOfScreen(screen)+10, // just in case
									0,
									CopyFromParent,
									InputOutput, 
									CopyFromParent,
									CWBackPixel | CWOverrideRedirect,
									&xswa))) 
	{
		XMapRaised(dsp, blackWindow);  

		/* grab the cursor */
		if (XGrabPointer(dsp, blackWindow, True, (unsigned int) AllPointerEventMask,
			   GrabModeAsync, GrabModeAsync, None, myCursor, CurrentTime) == GrabSuccess) 
		{
			/* grab the keyboard */
			if (XGrabKeyboard(dsp, blackWindow, True, GrabModeAsync, GrabModeAsync, CurrentTime)
					== GrabSuccess) 
			{
				pthread_mutex_lock ( & ( wr -> mutex ) ) ;
				wr -> ready = 1 ;
				pthread_mutex_unlock ( & ( wr -> mutex ) ) ;
				pthread_cond_broadcast ( & ( wr -> change ) ) ;

				/* We don't want this to block in XtProcessEvent
				 * so start a timer that ticks every 100ms.
				 * The event this generates will cause the quit flag
				 * to be checked again. */

				countdown = 200 ;		/* 200 x 100ms = 20s */
				(void) XtAddTimeOut ( 100 , x11_whitescreen_timer , & countdown ) ;
				while ( 1 )
				{
					pthread_mutex_lock ( & ( wr -> mutex ) ) ;
					if ( wr -> quit )
					{
						pthread_mutex_unlock ( & ( wr -> mutex ) ) ;
						break ;
					}
					if ( countdown < 0 )
					{
						wr -> timeout = 1 ;
						pthread_mutex_unlock ( & ( wr -> mutex ) ) ;
						break ;
					}
					pthread_mutex_unlock ( & ( wr -> mutex ) ) ;

					XtProcessEvent ( XtIMTimer | XtIMXEvent ) ;
				}

				// shutdown
				XUngrabKeyboard(dsp, CurrentTime);
			}
			XUngrabPointer(dsp, CurrentTime);
		}
		//  XUnmapWindow(dsp, myWindow); /* why is this commented out? */
		//  XDestroyWindow(dsp, myWindow);
	}
	XCloseDisplay(dsp);

	pthread_mutex_lock ( & ( wr -> mutex ) ) ;
	if ( ! wr -> ready )
	{
		wr -> error_occured = 1 ;
	}
	pthread_mutex_unlock ( & ( wr -> mutex ) ) ;
	pthread_cond_broadcast ( & ( wr -> change ) ) ;
	return NULL ;
}

/*
 * Overlay window functions
 */


static void * x11_window_loop(void* arg) ;

int x11_open_window ( dxr2_status_info_t * dxr2_info , geom_t geom )
{
	Dimension x, y;
	String		fake_argv [ 2 ] = { "Minidxr2 DVD" , NULL } ;
	int			fake_argc = 1 ;

	if ( dxr2_info -> window_open )
	{
		return 1 ;
	}
	if ( dxr2_info -> toplevel == NULL )
	{
		dxr2_info -> toplevel = XtAppInitialize(& dxr2_info -> application ,
						fake_argv [ 0 ] , NULL, 0,
						& fake_argc , fake_argv ,
						NULL,
						NULL, 0);
	}

	if ( ! XtIsWidget ( dxr2_info -> toplevel ) )
	{
		return 0 ;
	}

	/* copy the size so that it can be used from another thread without
	 * using more mutexes */
	dxr2_info -> max_x = dxr2_info->vgaBuf.xScreen ;
	dxr2_info -> max_y = dxr2_info->vgaBuf.yScreen ;
	dxr2_info -> window_open = 1 ;
	XtMakeResizeRequest( dxr2_info -> toplevel, geom.width, geom.height, &x, &y);
	XtRealizeWidget( dxr2_info -> toplevel);
	XMoveWindow( XtDisplay( dxr2_info -> toplevel), XtWindow( dxr2_info -> toplevel), geom.x, geom.y);
	XSetWindowBackground( XtDisplay( dxr2_info -> toplevel), 
				XtWindow( dxr2_info -> toplevel), 0x0000ff ); 

	dxr2_info -> quit_flag = 0 ;
	pthread_mutex_init ( & ( dxr2_info -> x_thread_mutex ) , NULL ) ;
	pthread_create ( & dxr2_info -> x_thread, NULL , x11_window_loop , dxr2_info ) ;
	return 1 ;
}

int x11_close_window ( dxr2_status_info_t * dxr2_info )
{
	if ( ! dxr2_info -> window_open )
	{
		return 0 ;
	}
	pthread_mutex_lock ( & ( dxr2_info -> x_thread_mutex ) ) ;
	dxr2_info -> quit_flag = 1 ;
	pthread_mutex_unlock ( & ( dxr2_info -> x_thread_mutex ) ) ;
	pthread_join ( dxr2_info -> x_thread , NULL ) ;
	pthread_mutex_destroy ( & ( dxr2_info -> x_thread_mutex ) ) ;
	dxr2_info -> window_open = 0 ;
	return 1 ;
}


static void enlarge_geom ( dxr2_status_info_t * dxr2_info , geom_t * geom , int margin )
{
	geom -> x -= margin ;
	geom -> y -= margin ;
	geom -> width += margin * 2 ;
	geom -> height += margin * 2 ;
	if ( geom -> x < 0 )
	{
		geom -> width += geom -> x ;
		geom -> x = 0 ;
	}
	if ( geom -> y < 0 )
	{
		geom -> height += geom -> y ;
		geom -> y = 0 ;
	}
	if ( geom -> x >= dxr2_info -> max_x )
	{
		geom -> width = ( dxr2_info -> max_x - geom -> x ) + margin ;
	}
	if ( geom -> y >= dxr2_info -> max_y )
	{
		geom -> height = ( dxr2_info -> max_y - geom -> y ) + margin ;
	}
	if ( geom -> width < 10 )
	{
		geom -> width = 10 ;
	}
	if ( geom -> height < 10 )
	{
		geom -> height = 10 ;
	}
	if ( geom -> x < 0 )
	{
		geom -> x = 0 ;
	}
	if ( geom -> y < 0 )
	{
		geom -> y = 0 ;
	}
}

static void x11_resize_handle ( dxr2_status_info_t * dxr2_info , Widget w , 
						XtPointer passed_data, XEvent* event, Boolean* bool)
{
	XWindowAttributes attrib;
	Position x, y ; /* , x1, y1; */
	geom_t geom;

	XGetWindowAttributes( XtDisplay(w), XtWindow(w) , &attrib );
	attrib.y = 0; // y is reported incorrectly
	attrib.x = 0; // reset x just incase
	XtTranslateCoords(w, attrib.x, attrib.y, &x, &y);

	geom.x=x;
	geom.y=y;
	geom.width=attrib.width;
	geom.height=attrib.height;

	enlarge_geom ( dxr2_info , & geom , 2 ) ;

	dxr2_conditional_view_change ( dxr2_info , geom ) ;

	/* attrib.x = 0;
	attrib.y = 0;
	XtTranslateCoords(w, attrib.x, attrib.y, &x1, &y1); */
}


static void x11_window_timer ( XtPointer p , XtIntervalId * interval )
{
	timer_data			* td ;

	td = (timer_data *) p ;
	td -> timer_id = XtAppAddTimeOut ( td -> app , 100 , 
								x11_window_timer , p ) ;
}

static void * x11_window_loop(void* arg)
{
	XEvent event;
	dxr2_status_info_t * dxr2_info = (dxr2_status_info_t *) arg ;
	int quit_flag = 0 ;
	int refresh = 1 ;
	timer_data			td ;

	XtRealizeWidget ( dxr2_info -> toplevel ) ;
	x11_resize_handle ( dxr2_info , dxr2_info -> toplevel , NULL , NULL , NULL ) ;
	td . app = dxr2_info -> application ;
	td . timer_id = XtAppAddTimeOut ( dxr2_info -> application , 100 , 
				x11_window_timer , & td ) ;
	while ( ! quit_flag )
	{ 
		if ( refresh )
		{
			x11_resize_handle ( dxr2_info , dxr2_info -> toplevel , NULL , NULL , NULL ) ;
			refresh = 0 ;
		}

		if ( XtAppPeekEvent ( dxr2_info -> application , & event ) )
		{
			XtAppProcessEvent ( dxr2_info -> application , XtIMXEvent | XtIMTimer ) ;
			if(event.type == ConfigureNotify )  /* aka. resize */
			{ 
				x11_resize_handle ( dxr2_info , dxr2_info -> toplevel , NULL , NULL , NULL ) ;
			}
			else if ( event.type == UnmapNotify ) /* minimise */
			{ 
				dxr2_hide_overlay ( dxr2_info ) ;
			} else if ( event.type == MapNotify )  /* restore */
			{
				x11_resize_handle ( dxr2_info , dxr2_info -> toplevel , NULL , NULL , NULL ) ;
			}
		} else {
			XtAppProcessEvent ( dxr2_info -> application , XtIMXEvent | XtIMTimer ) ;
		}

		pthread_mutex_lock ( & ( dxr2_info -> x_thread_mutex ) ) ;
		quit_flag = dxr2_info -> quit_flag ;
		if ( dxr2_info -> win_refresh_flag )
		{
			refresh = 1 ;
			dxr2_info -> win_refresh_flag = 0 ;
		}
		pthread_mutex_unlock ( & ( dxr2_info -> x_thread_mutex ) ) ;
	}
	XtRemoveTimeOut ( td . timer_id ) ;
	XtUnmapWidget ( dxr2_info -> toplevel ) ;
	XtUnrealizeWidget ( dxr2_info -> toplevel ) ;

	while ( XtAppPending ( dxr2_info -> application ) & XtIMXEvent )
	{
		XtAppProcessEvent ( dxr2_info -> application , XtIMXEvent ) ;
	}
	dxr2_hide_overlay ( dxr2_info ) ;
	return NULL ;
}

void x11_window_refresh ( dxr2_status_info_t * dxr2_info )
{
	pthread_mutex_lock ( & ( dxr2_info -> x_thread_mutex ) ) ;
	dxr2_info -> win_refresh_flag = 1 ;
	pthread_mutex_unlock ( & ( dxr2_info -> x_thread_mutex ) ) ;
}


#endif /* HAVE_X_OVERLAY_SUPPORT */
#endif /* HAVE_DXR2 */

