/*
 * ifo_parse.c
 *
 * IFO parsing functions using Thomas Mirlacher's libifo.
 *
 * Original version:
 *
 * Copyright (C) 1998-2000  Thomas Mirlacher
 *
 * Changes in this version:
 *
 * Copyright (C) 2001 Jack Whitham
 *
 * Subtitle colouring patch: XXX
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * The author may be reached as <jdw108@york.ac.uk>
 *
 *------------------------------------------------------------
 *
 */


#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <inttypes.h>
#include <string.h>
#include <assert.h>

#include "ifo/libifo/ifo.h"
// i've got link errors when i include bswap instead of byteswap.h -- does anybody knows why?
//#include <bswap.h>
#include <byteswap.h>

#include "ifo_parse.h"
#include "player.h"
#include "parser.h"



/* local functions */
static play_stack_cell * try_out_cell ( play_stack_cell * start , int position ) ;
static int find_next_part ( int * cell_num , int * cell_vob , 
			player_info_t * p , play_stack_cell ** cell ) ;
static ifo_t * get_ifo ( player_info_t * p , int title ) ;
static void free_playback_stack ( play_stack_node ** play_stack , int num_vobs ) ;
static int compute_playback_stack ( player_info_t * p , int title ) ;
static ifo_video_info_t * get_video_ptr ( player_info_t * p , int title ) ;
static ifo_audio_t * get_audio_ptr ( player_info_t * p , int title , int stream ) ;
static int get_first_time_map_table ( ifo_t * the_ifo , int * timeunit ,
					uint32_t ** start_ptr , uint32_t ** end_ptr ) ;
static int playback_stack_check ( player_info_t * p , int title ) ;

/* Data follower.
 * This function is intended to be given a current position (which may be invalid).
 * From this, it decides where the next position should be.
 */

int ifo_data_follower ( player_info_t * p , 
			int * position , int * until , int * title ,
			int desired_angle , int mode )
{
	play_stack_cell		* cell = NULL ;
	int					i , j ;
	int					cell_num = 0 ;
	int					cell_vob = 0 ;
	int					rc ;
	int					confused = 0 ;
	int					eot = 0 ;

	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;
	if ( ! p -> ifo_available )
	{
		/* use the non-ifo (linear) data follower -
		 * tell it to play 128 more chunks from this position. */

		(* until) = (* position) + ( CHUNK_SIZE * 128 ) ;
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_OK ;
	}

	rc = playback_stack_check ( p , (* title) ) ;
	if ( rc != IFO_OK )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return rc ;
	}

	if (( p -> play_stack_num_vobs < 1 )
	|| ( p -> play_stack [ 0 ] -> num_cells < 1 ))
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_SILLY_DATA ;
	}


	if (( p -> play_stack_accelerator_cell >= 0 )
	&& ( p -> play_stack_accelerator_vob >= 0 )
	&& ( p -> play_stack_accelerator_vob < 
				p -> play_stack_num_vobs )
	&& ( p -> play_stack_accelerator_cell < 
				p -> play_stack [ p -> play_stack_accelerator_vob ] -> 
					num_cells ))
	{
		cell = try_out_cell ( 
					p -> play_stack [ p -> play_stack_accelerator_vob ] -> 
						cells [ p -> play_stack_accelerator_cell ] ,
					(* position) ) ;
		cell_num = p -> play_stack_accelerator_cell ;
		cell_vob = p -> play_stack_accelerator_vob ;
	} 

	if ( cell == NULL )
	{
		/* where is position in the stack? */
		for ( i = 0 ; i < p -> play_stack_num_vobs ; i ++ )
		{
			for ( j = 0 ; j < p -> play_stack [ i ] -> num_cells ; j ++ )
			{
				if (( cell = try_out_cell ( p -> play_stack [ i ] -> cells [ j ] ,
						(* position ))) != NULL )
				{
					cell_num = j ;
					break ;
				}
			}

			if ( cell != NULL )
			{
				cell_vob = i ;
				break ;
			}
		}
	}

	if ( cell == NULL )
	{
		cell = p -> play_stack [ 0 ] -> cells [ 0 ] ;
		confused = 1 ;
	}

	if ( mode == IFO_DFMODE_FIND_NEXT )
	{
		/* skip multi angles */
		while (( cell -> next != NULL )
		&& ( cell -> start == cell -> next -> start ))
		{
			/* I don't think this will ever happen. */
			printf ( "Look out! The Spanish Inquisition!\n" ) ;
			cell = cell -> next ;
		}

		cell = cell -> next ;

		if ( cell == NULL )
		{
			eot = find_next_part ( & cell_num , & cell_vob , p , & cell ) ;

			if ( cell -> next != NULL )
			{
				if ( p -> play_skip_angles == 0 )
				{
					p -> play_skip_angles = desired_angle ; /* Angle required */
					cell_num += p -> play_skip_angles - 1 ;
					if ( cell_num >= p -> play_stack [ cell_vob ] -> num_cells )
					{
						eot = find_next_part ( & cell_num , & cell_vob , p , & cell ) ;
					} else {
						cell = p -> play_stack [ cell_vob ] -> cells [ cell_num ] ;
					}
				} else {
					cell_num += p -> play_num_angles - p -> play_skip_angles ;
					if ( cell_num >= p -> play_stack [ cell_vob ] -> num_cells )
					{
						eot = find_next_part ( & cell_num , & cell_vob , p , & cell ) ;
					} else {
						cell = p -> play_stack [ cell_vob ] -> cells [ cell_num ] ;
					}
					p -> play_skip_angles = 0 ;
				}
			}
		}

		assert ( cell != NULL ) ;

		/* "The Matrix Fix".
		 * Don't replay a cell that's just been played. */
		if (( cell -> next != NULL )
		&& ((* position) > cell -> start ))
		{
			cell = cell -> next ;
		}

		(* position) = cell -> start ;
	} else if ( mode == IFO_DFMODE_FIND_BEGINNING )
	{
		(* position) = cell -> start ;
	} /* else, don't change the position */

	(* until) = cell -> end ;
	p -> play_stack_accelerator_vob = cell_vob ;
	p -> play_stack_accelerator_cell = cell_num ;

	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
	if ( eot )
	{
		return IFO_END_OF_TITLE ;
	} else {
		/* CONFUSED is not actually an error. It indicates that
		 * the data follower was unable to find the given position in the
		 * file and therefore just picked one (the beginning) */
		return confused ? IFO_CONFUSED : IFO_OK ;
	}
}


/*
 * ifo_recompute_tables
 *
 * Tells the IFO parser that the tables will need to be
 * recomputed (because the input files have changed or something)
 */

void ifo_recompute_tables ( player_info_t * p )
{
	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;

	if ( ! p -> ifo_available )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return ;
	}

	free_playback_stack ( p -> play_stack , p -> play_stack_num_vobs ) ;
	p -> play_stack_setup = 0 ;
	p -> play_stack = NULL ;
	p -> play_stack_num_vobs = 0 ;
	p -> play_stack_accelerator_cell =
		p -> play_stack_accelerator_vob = -1 ;

	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
}



/*
 * ifo_get_num_chapters
 *
 * Returns the number of chapters on the given title
 *
 */

int ifo_get_num_chapters ( player_info_t * p , int title )
{
	int		rc ;

	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;

	rc = playback_stack_check ( p , title ) ;
	if ( rc != IFO_OK )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return rc ;
	}

	rc = p -> play_stack_num_vobs ;

	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;

	return rc ;
}


/*
 * ifo_get_chapter_position
 *
 * Return the position of the start of the given chapter.
 */

int ifo_get_chapter_position ( player_info_t * p , int title , int chapter )
{
	int		rc , out ;

	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;

	rc = playback_stack_check ( p , title ) ;
	if ( rc != IFO_OK )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return rc ;
	}

	if (( chapter <= p -> play_stack_num_vobs )
	&& ( chapter > 0 ))
	{
		out = p -> play_stack [ chapter - 1 ] -> cells [ 0 ] -> start ;
	} else {
		out = IFO_SILLY_PARAMETERS ;
	}
	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;

	return out ;
}

/*
 * ifo_available
 *
 * Returns 1 if proper IFO parsing is available.
 */

int ifo_available ( player_info_t * player_info )
{
	int			rc ;

	pthread_mutex_lock ( & ( player_info -> ifo_mutex ) ) ;
	rc = player_info -> ifo_available ;
	pthread_mutex_unlock ( & ( player_info -> ifo_mutex ) ) ;

	return rc ;
}

/* 
 * ifo_open
 *
 * Opens the disc IFO and any titles that may exist. 
 */

int ifo_open ( player_info_t * p )
{
	int						title ;

	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;

	if (( p -> disc_ifo = ifo_examine_title ( p , 0 , & p -> disc_ifo_fd )) == NULL )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return 0 ;
	}
	for ( title = 1 ; title < 100 ; title ++ )
	{
		if (( p -> title_ifo [ title ] = ifo_examine_title ( p , 
							title , & p -> title_ifo_fd [ title ] )) == NULL )
		{
			break ;
		}
	}
	p -> num_physical_titles = title - 1 ;
	p -> ifo_available = 1 ;

	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
	return 1 ;
}


/*
 * ifo_close
 *
 * Closes the IFO
 *
 */

int ifo_close ( player_info_t * p )
{
	int		title ;

	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;
	if ( ! p -> ifo_available )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NOT_AVAILABLE ;
	}

	for ( title = 1 ; title < ( p -> num_physical_titles + 1 ) ; title ++ )
	{
		ifoClose ( p -> title_ifo [ title ] ) ;
		close ( p -> title_ifo_fd [ title ] ) ;
	}
	ifoClose ( p -> disc_ifo ) ;
	close ( p -> disc_ifo_fd ) ;
	p -> ifo_available = 0 ;

	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
	return 1 ;
}


int ifo_get_num_sub_streams ( player_info_t * p , int title )
{
	uint8_t		* table_ptr ;
	uint8_t 	* sub_ptr = NULL ;
	int			rc ;

	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;
	if ( ! p -> ifo_available )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NOT_AVAILABLE ;
	}
	if ( ifoIsVTS ( get_ifo ( p , title ) ) ) 
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return 0 ;
	}
	table_ptr = (uint8_t *) ( get_ifo ( p , title ) -> tbl [ ID_MAT ] + IFO_OFFSET_SUBPIC ) ;
	rc = ifoGetSPU ( table_ptr , & sub_ptr ) ;
	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;

	if ( sub_ptr != NULL )
	{
#ifdef IFO_DEBUGGING
		printf ( "num sub streams = %d\n" , rc ) ;
#endif
		return rc ;
	} else {
		return 0 ;
	}
}

const clut_t *ifo_get_sub_clut ( player_info_t * p , int title )
{
       int                     rc ;
       const clut_t            *clut ;
	   ifo_t				   * ifo ;
       pgc_t			       * pgci ;

       pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;
       if ( ! p -> ifo_available )
       {
               pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
               return NULL ;
       }
	   ifo = get_ifo ( p , title ) ;

       if (( ifo == NULL )
	   || ( ifoIsVTS ( ifo ) ))
       {
               pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
               return NULL ;
       }

	   pgci = (pgc_t *) ifoGetPGCI( ifo , IFO_TITLE , 0 ) ;
	   assert ( pgci != NULL ) ;

       clut = ifoGetCLUT( pgci ) ;
       pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
       return clut;
}

int ifo_settings_sub_langcode ( player_info_t * p , 
		int title , int stream , int * langcode )
{
	uint8_t		* table_ptr ;
	uint8_t 	* ptr ;
	ifo_spu_t	* spu_ptr ;
	int			num ;
	int			ok = 0 ;

	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;
	if ( ! p -> ifo_available )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NOT_AVAILABLE ;
	}
	if ( ifoIsVTS ( get_ifo ( p , title ) ) ) 
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NOT_AVAILABLE ;
	}
	table_ptr = (uint8_t *) ( get_ifo ( p , title ) -> tbl [ ID_MAT ] + IFO_OFFSET_SUBPIC ) ;
	num = ifoGetSPU ( table_ptr , & ptr ) ;

	if (( ptr != NULL )
	&& ( stream >= 0 )
	&& ( stream < num ))
	{
		spu_ptr = (ifo_spu_t *) ( ptr + ( stream * sizeof ( ifo_spu_t ) )) ;
		(* langcode) = spu_ptr -> lang_code ;
		ok = spu_ptr -> caption ;
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
	} else {
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NOT_AVAILABLE ;
	}

	switch ( ok )
	{
		case 0 :
		case 1 :		return IFO_LANGCODE_NORMAL ;
		case 3 :		return IFO_LANGCODE_COMMENTARY ;
		default :		
#ifdef IFO_DEBUGGING
						printf ( "Unknown type %d\n" , ok ) ;
#endif
						return IFO_LANGCODE_UNKNOWN ;
	}
}

int ifo_get_num_audio_streams ( player_info_t * p , int title )
{
	uint8_t		* table_ptr ;
	uint8_t 	* audio_ptr ;
	int			rc ;

	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;
	if ( ! p -> ifo_available )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NOT_AVAILABLE ;
	}
	if ( ifoIsVTS ( get_ifo ( p , title ) ) ) 
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return 0 ;
	}
	table_ptr = (uint8_t *) ( get_ifo ( p , title ) -> tbl [ ID_MAT ] + IFO_OFFSET_AUDIO ) ;
	rc = ifoGetAudio ( table_ptr , & audio_ptr ) ;
	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;

	return rc ;
}

/*
 * ifo_settings...
 *
 * These functions return the best video/audio settings for the specified title.
 */

int ifo_settings_aspect_16_9 ( player_info_t * p , int title )
{
	ifo_video_info_t *video ;
	
	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;
	if ( ! p -> ifo_available )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NOT_AVAILABLE ;
	}
	video = get_video_ptr ( p , title ) ;
	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;

	return ( video -> ratio == 3 ) ;
}

int ifo_settings_pan_scan ( player_info_t * p , int title )
{
	ifo_video_info_t *video ;
	
	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;
	if ( ! p -> ifo_available )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NOT_AVAILABLE ;
	}
	video = get_video_ptr ( p , title ) ;
	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;

	return ( video -> perm_displ == 1 ) ;
}

int ifo_settings_langcode ( player_info_t * p , int title , int stream )
{
	ifo_audio_t		* audio_ptr ;

	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;
	if ( ! p -> ifo_available )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NOT_AVAILABLE ;
	}
	audio_ptr = get_audio_ptr ( p , title , stream ) ;
	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;

	if ( audio_ptr != NULL )
	{
		return audio_ptr -> lang_code ;
	} else {
		return IFO_UNKNOWN ;
	}
}

int ifo_settings_audio_format ( player_info_t * p , int title , int stream )
{
	ifo_audio_t		* audio_ptr ;

	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;
	if ( ! p -> ifo_available )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NOT_AVAILABLE ;
	}
	audio_ptr = get_audio_ptr ( p , title , stream ) ;
	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;

	if ( audio_ptr != NULL )
	{
		switch ( audio_ptr -> coding_mode & 0x07 )
		{
			case 0 :		return OUTPUT_AUDIO_AC3 ;
			case 2 :		return OUTPUT_AUDIO_MPEG ;
			case 3 :		return OUTPUT_AUDIO_MPEG ; /* really MPEG2ext */
			case 4 :		return OUTPUT_AUDIO_LPCM ;
			case 5 :		return OUTPUT_AUDIO_AC3 ; /* really DTS */
			default :		return IFO_UNKNOWN ;
		}
	} else {
		return IFO_UNKNOWN ;
	}
}

int ifo_settings_audio_freq ( player_info_t * p , int title , int stream )
{
	ifo_audio_t		* audio_ptr ;

	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;
	if ( ! p -> ifo_available )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NOT_AVAILABLE ;
	}
	audio_ptr = get_audio_ptr ( p , title , stream ) ;
	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;

	if ( audio_ptr != NULL )
	{
		return ( audio_ptr -> sample_freq ? OUTPUT_AUDIO_FREQ_96 : OUTPUT_AUDIO_FREQ_48 ) ;
	} else {
		return IFO_UNKNOWN ;
	}
}

int ifo_settings_audio_width ( player_info_t * p , int title , int stream )
{
	ifo_audio_t		* audio_ptr ;

	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;
	if ( ! p -> ifo_available )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NOT_AVAILABLE ;
	}
	audio_ptr = get_audio_ptr ( p , title , stream ) ;
	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;

	if ( audio_ptr != NULL )
	{
		switch ( audio_ptr -> quantization )
		{
			case 0 :	return OUTPUT_AUDIO_WIDTH_16 ;
			case 1 :	return OUTPUT_AUDIO_WIDTH_20 ;
			case 2 :	return OUTPUT_AUDIO_WIDTH_24 ;
			default :	return OUTPUT_AUDIO_WIDTH_16 ;
		}
	} else {
		return IFO_UNKNOWN ;
	}
}

int ifo_settings_num_channels ( player_info_t * p , int title , int stream )
{
	ifo_audio_t		* audio_ptr ;

	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;
	if ( ! p -> ifo_available )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NOT_AVAILABLE ;
	}
	audio_ptr = get_audio_ptr ( p , title , stream ) ;
	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;

	if ( audio_ptr != NULL )
	{
		return audio_ptr -> num_channels + 1 ;
	} else {
		return IFO_UNKNOWN ;
	}
}

int ifo_get_position_for_time ( player_info_t * p , int title , int hour , int min , int second )
{
	uint32_t		* start_ptr ;
	uint32_t		* ptr ;
	uint32_t		* end_ptr ;
	int				timeunit ;
	float			seektime ;
	int				rc ;
	uint32_t		before_data ;
	uint32_t		after_data ;
	ifo_t			* the_ifo ;

	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;
	if ( ! p -> ifo_available )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NOT_AVAILABLE ;
	}
	the_ifo = get_ifo ( p , title ) ;
	rc = get_first_time_map_table ( the_ifo , & timeunit , & start_ptr , & end_ptr ) ;
	if ( rc != IFO_OK )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return rc ;
	}

	seektime = ( (float) second + ( min * 60 ) + ( hour * 3600 )) / (float) timeunit ;
	if ( seektime < 0.0 )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_SILLY_PARAMETERS ;
	}
	if ( seektime < 1.0 )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return 0 ;
	}

	ptr = start_ptr ;
	ptr += ( (int) seektime - 1 ) ;
	if (( ptr >= start_ptr )
	&& ( ptr < ( end_ptr - 1 )))
	{
		before_data = be2me_32 ( (* ptr) ) & 0x7fffffff ;
		ptr ++ ;
		after_data = be2me_32 ( (* ptr) ) & 0x7fffffff ;
		if ( before_data <= after_data )
		{
			seektime -= (float) ( (int) seektime ) ;	/* get fractional part of seektime */
			rc = ( (int) (( (float) ( after_data - before_data )) * seektime )
						+ before_data ) ;
			if ( rc >= 0 )
			{
				pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
				return rc ;
			} else {
				pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
				return IFO_SILLY_DATA ;
			}
		} else {
			pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
			return IFO_SILLY_DATA ;
		}
	} else {
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_SILLY_PARAMETERS ;
	}
}

int ifo_get_time_for_position ( player_info_t * p , int title , int position ,
		int * hour , int * min , int * second )
{
	ifo_t			* the_ifo ;
	int				rc ;

	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;
	if ( ! p -> ifo_available )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NOT_AVAILABLE ;
	}
	the_ifo = get_ifo ( p , title ) ;

	rc = ifo_direct_get_time_for_position ( the_ifo , position , hour , min , second ) ;

	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
	return rc ;
}

int ifo_direct_get_time_for_position ( ifo_t * the_ifo , int position , int * hour , int * min , int * second )
{
	uint32_t		* start_ptr ;
	uint32_t		* ptr ;
	uint32_t		* end_ptr ;
	int				timeunit ;
	int				rc ;
	uint32_t		data , lastdata ;
	int				time , lasttime ;

	rc = get_first_time_map_table ( the_ifo , & timeunit , & start_ptr , & end_ptr ) ;
	if ( rc != IFO_OK )
	{
		return rc ;
	}

	ptr = start_ptr ;
	data = lastdata = lasttime = 0 ;
	while ( ptr < end_ptr )		/* how will this handle multi-angle discs? */
	{
		data = be2me_32 ( (* ptr) ) & 0x7fffffff ;
		ptr ++ ;
		if ( data < lastdata )
		{
			break ;
		} else {
			if (( lastdata <= position )
			&& ( position < data ))
			{
				/* position is between lastdata and data */
				time = (int) ( (((float) ( position - lastdata ) 
									/ (float) ( data - lastdata )) *
							(float) timeunit ) + (float) lasttime ) ;
				(* second) = time % 60 ;
				(* min) = ( time / 60 ) % 60 ;
				(* hour) = time / 3600 ;
#ifdef IFO_DEBUGGING
				printf ( "position %d located at %d:%d:%d\n" ,
							position , (* hour) , (* min) , (* second) ) ;

#endif
				return IFO_OK ;
			} else {
				lastdata = data ;
				lasttime += timeunit ;
			}
		}
	}
	/* sometimes, we're looking for the last time in the file.
	 * In this case, the position is either specified as less
	 * than zero, or as greater than the last place in the file.
	 * Need to decide if the position is 'close enough' */
	if (( data > ( position - FRAME_SIZE ))
	|| ( position < 0 ))
	{
		/* close enough. Yes, this is quite a hack! */
		(* second) = lasttime % 60 ;
		(* min) = ( lasttime / 60 ) % 60 ;
		(* hour) = lasttime / 3600 ;
#ifdef IFO_DEBUGGING
		printf ( "END position %d located at %d:%d:%d\n" ,
					position , (* hour) , (* min) , (* second) ) ;
#endif
		return IFO_OK ;
	} else {
		return IFO_UNKNOWN ;
	}
}


int ifo_get_pos_from_main_title_search_pointer ( player_info_t * p , int map_number ,
			int * title , int * angle )
{
	/* This was called toast in tlassen's dxr2 program. As good a name as any.. :) */

	uint8_t			* ptr ;
	ifo_hdr_t		* header ;
	ifo_tsp_t		* tsp ;
	ifo_t			* ifo ;

	pthread_mutex_lock ( & ( p -> ifo_mutex ) ) ;
	if ( ! p -> ifo_available )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NOT_AVAILABLE ;
	}
	ifo = p -> disc_ifo ;
	if ( ifo == NULL )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_NO_TABLE ;
	}

	ptr = ifo->tbl[ID_PTT_SRPT];
	header = (ifo_hdr_t *) ptr ;

	if ( map_number >= be2me_16 ( header -> num ) )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_OFF_END_OF_FILE ;
	}

	ptr += sizeof ( ifo_hdr_t ) + ( map_number * sizeof ( ifo_tsp_t ) ) ;
	if ( ptr > ( (uint8_t *) header + be2me_32 ( header -> len ) + sizeof ( ifo_tsp_t )) )
	{
		pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
		return IFO_SILLY_DATA ;
	}

	tsp = (ifo_tsp_t *) ptr ;

	(* title) = tsp -> title_set_num ;		/* physical title number? */
	(* angle) = tsp -> num_angles ;		/* is this right?? */

	pthread_mutex_unlock ( & ( p -> ifo_mutex ) ) ;
	return IFO_OK ;
}

void ifo_setup ( player_info_t * player_info )
{
	pthread_mutex_init(&(player_info->ifo_mutex), NULL);

	player_info -> num_physical_titles = 0 ;
	player_info -> play_stack = NULL ;
}





static ifo_t * get_ifo ( player_info_t * p , int title )
{
	if ( ! p -> ifo_available )
	{
		return NULL ;
	}

	if ( title > p -> num_physical_titles )
	{
		return NULL ;
	}
	if ( title > 0 )
	{
		return ( p -> title_ifo [ title ] ) ;
	} else {
		return ( p -> disc_ifo ) ;
	}
}


static int find_next_part ( int * cell_num , int * cell_vob , 
			player_info_t * p , play_stack_cell ** cell )
{
	(* cell_num) ++ ;
	if ( (* cell_num) < p -> play_stack [ (* cell_vob) ] -> num_cells )
	{
		(* cell) = p -> play_stack [ (* cell_vob) ] -> cells [ (* cell_num) ] ;
	} else {
		(* cell_num) = 0 ;
		(* cell_vob) ++ ;
		if ( (* cell_vob) < p -> play_stack_num_vobs )
		{
			(* cell) = p -> play_stack [ (* cell_vob) ] -> cells [ 0 ] ;

			/* XXX jump some cells that aren't the right angle */
		} else {
			/* Change title? */
			(* cell) = p -> play_stack [ 0 ] -> cells [ 0 ] ;
			(* cell_num) = (* cell_vob) = 0 ;
			return 1 ;
		}
	}
	assert ( (* cell) != NULL ) ;
	return 0 ;
}

static play_stack_cell * try_out_cell ( play_stack_cell * start , int position )
{
	play_stack_cell		* current ;

	current = start ;
	while ( current != NULL )
	{
		if (( current -> start <= position )
		&& ( current -> end >= position ))
		{
			return ( current ) ;
		}
		current = current -> next ;
	}
	return NULL ;
}

static void free_playback_stack ( play_stack_node ** play_stack , int num_vobs )
{
	int					i , j ;
	play_stack_cell		* next ;
	play_stack_cell		* current ;

	for ( i = 0 ; i < num_vobs ; i ++ )
	{
		if ( play_stack [ i ] != NULL )
		{
			for ( j = 0 ; j < play_stack [ i ] -> num_cells ; j ++ )
			{
				current = play_stack [ i ] -> cells [ j ] ;
				while ( current != NULL )
				{
					next = current -> next ;
					free ( current ) ;
					current = next ;
				}
			}
			free ( play_stack [ i ] -> cells ) ;
			free ( play_stack [ i ] ) ;
		}
	}
	free ( play_stack ) ;
}

static int compute_playback_stack ( player_info_t * p , int title )
{
	uint8_t			* ptr ;
	uint8_t			* continue_ptr ;
	uint8_t			* endptr ;
	ifo_hdr_t		* header ;
	ifo_t			* ifo ;
	ifo_caddr_t		* cell ;
	ifo_caddr_t		* prev_cell ;
	ifo_tsp_t		* tsp ;
	ifo_ptt_t		* ptt ;
	ifo_ptt_sub_t	* ptt_title ;
	pgc_t			* pgc_ptr ;
	uint8_t			* program_map ;
	ifo_pgc_cpos_t	* cell_pos ;
	int				i , j , current_vob_id , tt , max_cells ;
	int				length , num_vobs , current_vob , current_cell_id ;
	int				total_num_cells , num_cells_directly_in_stack ;
	int				title_loc = 0 ;
	int				chapter , program_map_num , cell_pos_num , end_num ;
	int				stack_vob_num , stack_cell_num , start_num , curcell ;
	play_stack_cell	* current_cell_ptr ;
	play_stack_node ** play_stack ;
	play_stack_node ** play_stack_2 ;

	/* work out the number of angles present in this title */
	ifo = p -> disc_ifo ;
	if ( ifo == NULL )
	{
		return IFO_NO_TABLE ;
	}

	ptr = ifo -> tbl [ ID_PTT_SRPT ] ;
	header = (ifo_hdr_t *) ptr ;

	tt = be2me_16 ( header -> num ) ;

	p -> play_num_angles = 1 ;

	for ( i = 0 ; i < tt ; i ++ )
	{
		ptr += sizeof ( ifo_hdr_t ) + ( i * sizeof ( ifo_tsp_t ) ) ;
		if ( ptr > ( (uint8_t *) header + be2me_32 ( header -> len ) + sizeof ( ifo_tsp_t )) )
		{
			return IFO_SILLY_DATA ;
		}

		tsp = (ifo_tsp_t *) ptr ;

		if ( title == tsp -> title_set_num )
		{
			p -> play_num_angles = tsp -> num_angles ;
			title_loc = tsp -> vts_ttn - 1 ;
			break ;
		}
	}

	/* create the playback stack */
	ifo = get_ifo ( p , title ) ;
	if ( ifo == NULL )
	{
		return IFO_NO_TABLE ;
	}

	ptr = ifo -> tbl [ ID_TITLE_CELL_ADDR ] ;
	if ( ptr == NULL )
	{
		return IFO_NO_TABLE ;
	}

	if ( p -> play_stack_setup )
	{
		free_playback_stack ( p -> play_stack , p -> play_stack_num_vobs ) ;
		p -> play_stack_setup = 0 ;
		p -> play_stack = NULL ;
	}

	p -> play_stack_num_vobs = 0 ;

	header = (ifo_hdr_t *) ptr ;

	num_vobs = be2me_16 ( header -> num ) ;
	length = be2me_32 ( header -> len ) / sizeof ( ifo_caddr_t ) ;
	if ( length < 1 )
	{
		return IFO_SILLY_DATA ;
	}
	endptr = ptr + be2me_32 ( header -> len ) ; /* - sizeof ( ifo_caddr_t ) ;*/
	ptr += sizeof ( ifo_hdr_t ) ;

	cell = (ifo_caddr_t *) ptr ;

	play_stack = calloc ( num_vobs , sizeof ( play_stack_node ** ) ) ;
	current_vob = 0 ;

	for ( i = 0 ; ( i < length )
			&& ( (uint8_t *) cell < endptr ) ; i ++ )
	{
		assert ( current_vob < num_vobs ) ; 
	/* for ( i = 0 ; ( i < length )
			&& ( (uint8_t *) cell < endptr ) 
			&& ( current_vob < num_vobs ) ; i ++ )
	{ */

		prev_cell = cell = (ifo_caddr_t *) ptr ;
		assert ( (uint8_t *) cell < endptr ) ;
		current_vob_id = current_vob + 1 ;
					
		for ( j = 1 ; j < ( length + 1 - i ) ; j ++ )
		{
			prev_cell = cell ;
			cell = (ifo_caddr_t *) ( ptr + ( j * sizeof ( ifo_caddr_t ) )) ;
			if ( ! ( (uint8_t *) cell < endptr ) )
			{
				break ;
			}
			if ( be2me_16 ( cell -> vob_id ) != current_vob_id )
			{
				break ;
			}
		}
		continue_ptr = (uint8_t *) cell ;

		num_cells_directly_in_stack = prev_cell -> cell_id ;
		assert ( num_cells_directly_in_stack > 0 ) ;
		total_num_cells = j ;
		play_stack [ current_vob ] = 
				calloc ( 1 , sizeof ( play_stack_node ) ) ;
		play_stack [ current_vob ] -> num_cells = num_cells_directly_in_stack ;
#ifdef IFO_DEBUGGING 
		printf ( "HER: %i,%i\n" , current_vob ,
					num_cells_directly_in_stack ) ;
#endif
		play_stack [ current_vob ] -> cells = 
				calloc ( num_cells_directly_in_stack , sizeof ( play_stack_cell * ) ) ;

		cell = (ifo_caddr_t *) ptr ;

		for ( j = 0 ; ( j < num_cells_directly_in_stack )
						&& ( (uint8_t *) cell < endptr ) ; j ++ )
		{
			/* create an item in the stack for this first cell */
			cell = (ifo_caddr_t *) ptr ;
			current_cell_id = cell -> cell_id ;

			play_stack [ current_vob ] -> cells [ j ] =
					calloc ( 1 , sizeof ( play_stack_cell ) ) ;
			current_cell_ptr = play_stack [ current_vob ] -> cells [ j ] ;

			current_cell_ptr -> start = be2me_32 ( cell -> start ) ;
			current_cell_ptr -> end = be2me_32 ( cell -> end ) ;
			current_cell_ptr -> cell_id = current_cell_id ;
			current_cell_ptr -> next = NULL ;

			/* any more cells sharing this cell id? Add them to it */
			ptr += sizeof ( ifo_caddr_t ) ;
			cell = (ifo_caddr_t *) ptr ;

			while (( (uint8_t *) cell < endptr )
			&& ( current_cell_id == cell -> cell_id )
			&& ( current_vob_id == be2me_16 ( cell -> vob_id )))
			{
				current_cell_ptr -> next = calloc ( 1 , sizeof ( play_stack_cell ) ) ;
				current_cell_ptr = current_cell_ptr -> next ;

				current_cell_ptr -> start = be2me_32 ( cell -> start ) ;
				current_cell_ptr -> end = be2me_32 ( cell -> end ) + 1 ;
				current_cell_ptr -> cell_id = current_cell_id ;
				current_cell_ptr -> next = NULL ;

				ptr += sizeof ( ifo_caddr_t ) ;
				cell = (ifo_caddr_t *) ptr ;
			}
		}

		assert ( continue_ptr == ptr ) ;

		current_vob ++ ;
	}

	num_vobs = current_vob ;

	/* find the title in the PTT */
	/* This IFO thing is such a great standard, huh?
	 * I think we should have MORE obfuscated proprietary standards! */

	ptt = ifoGetPTT ( ifo ) ;

	/* ptt numbers are already in the correct byte-order */

	assert ( ptt != NULL ) ;

	ptt_title = ptt -> title + title_loc ;

	assert ( ptt_title -> num > 0 ) ;

	play_stack_2 = calloc ( ptt_title -> num , sizeof ( play_stack_node ** ) ) ;

	for ( chapter = 0 ; chapter < ptt_title -> num ; chapter ++ )
	{
		/* "alloc" */
		pgc_ptr = (pgc_t *) ifoGetPGCI ( ifo , IFO_TITLE ,
				ptt_title -> data [ chapter ] . pgc - 1 ) ;
		assert ( pgc_ptr != NULL ) ;

		program_map = (uint8_t *) ifoGetProgramMap ( pgc_ptr ) ;
		cell_pos = ifoGetCellPos ( pgc_ptr ) ;
		program_map_num = ifoGetProgramMapNum ( pgc_ptr ) ;
		cell_pos_num = ifoGetCellPosNum ( pgc_ptr ) ;

		play_stack_2 [ chapter ] =
				calloc ( 1 , sizeof ( play_stack_node ) ) ;
			
		/* Am worried about pg & cell_pos_num. assert here? */
		if ( ptt_title -> data [ chapter ] . pg < program_map_num )
		{
			end_num = program_map [ ptt_title -> data [ chapter ] . pg ] ;
			max_cells = play_stack_2 [ chapter ] -> num_cells = 
				program_map [ ptt_title -> data [ chapter ] . pg ] -
					program_map [ ptt_title -> data [ chapter ] . pg - 1 ] + 1 ;
		} else {
			end_num = cell_pos_num ;
			max_cells = play_stack_2 [ chapter ] -> num_cells = 
				cell_pos_num - 
					program_map [ ptt_title -> data [ chapter ] . pg - 1 ] + 1 ;
		}
		assert ( max_cells > 0 ) ;

		play_stack_2 [ chapter ] -> cells =
				calloc ( max_cells , sizeof ( play_stack_cell * ) ) ;
		play_stack_2 [ chapter ] -> cells [ 0 ] =
				calloc ( 1 , sizeof ( play_stack_cell ) ) ;

		/* "Initialice" */

		/* get2bytes ( buf )
		 * {
		 *     return bswap_16 (*((u_int16_t *)buf));
		 *     }
		 */

		ptr = (uint8_t *) cell_pos ;
		ptr += ( program_map [ ptt_title -> data [ chapter ] . pg - 1 ] - 1 )
				* sizeof ( uint32_t ) ;

		/* assert about ptr here? */

		stack_vob_num = be2me_16 ( (* ((uint16_t *) ptr)) ) - 1 ;
		ptr += sizeof ( uint16_t ) ;
		stack_cell_num = be2me_16 ( (* ((uint16_t *) ptr)) ) - 1 ; /* already swapped */

		assert ( stack_vob_num < num_vobs ) ;

		assert ( stack_cell_num <
			play_stack [ stack_vob_num ] -> num_cells ) ;

		memcpy ( play_stack_2 [ chapter ] -> cells [ 0 ] ,
				play_stack [ stack_vob_num ] -> cells [ stack_cell_num ] ,
				sizeof ( play_stack_cell ) ) ;
		play_stack [ stack_vob_num ] -> cells [ stack_cell_num ] -> next = NULL ;
		play_stack_2 [ chapter ] -> num_cells = 1 ;

		start_num = program_map [ ptt_title -> data [ chapter ] . pg - 1 ] ;
		curcell = 1 ;	
		if ( curcell < max_cells )
		{
			while ( ( start_num + 1 ) < end_num )
			{
				assert ( curcell < max_cells ) ;
				if (( play_stack_2 [ chapter ] -> cells [ curcell ] != NULL )
				&& ( play_stack_2 [ chapter ] -> cells [ curcell ] -> next != NULL ))
				{
					stack_cell_num ++ ;
					while ( curcell < p -> play_num_angles )
					{
						/* Not convinced about the next line.
						 * We may need the num_cell thing really */
						if ( stack_cell_num < 
									play_stack [ stack_vob_num ] -> num_cells )
						{
							play_stack_2 [ chapter ] -> cells [ curcell ] =
									calloc ( 1 , sizeof ( play_stack_cell ) ) ;
							memcpy ( play_stack_2 [ chapter ] -> 
										cells [ curcell ] ,
									play_stack [ stack_vob_num ] -> 
										cells [ stack_cell_num ] ,
									sizeof ( play_stack_cell ) ) ;
							play_stack [ stack_vob_num ] -> 
									cells [ stack_cell_num ] -> next = NULL ;
							play_stack_2 [ chapter ] -> num_cells ++ ;
							stack_cell_num ++ ;
							/* assert on stack_cell_num ? */
						} else {
							stack_cell_num = 0 ;
							stack_vob_num ++ ;
							/* assert on stack_vob_num ? */
						}
						curcell ++ ;
						assert ( curcell < max_cells ) ;
						start_num ++ ;
					}
				} else {
					start_num ++ ;
					cell_pos = ifoGetCellPos ( pgc_ptr ) ;
					ptr = (uint8_t *) cell_pos ;
					ptr += ( start_num - 1 ) *
							sizeof ( uint32_t ) ;
					stack_vob_num = be2me_16 ( (* ((uint16_t *) ptr)) ) - 1 ;
					ptr += sizeof ( uint16_t ) ;
					stack_cell_num = be2me_16 ( (* ((uint16_t *) ptr)) ) - 1 ;

					assert ( stack_vob_num < num_vobs ) ;
					assert ( stack_cell_num <
						play_stack [ stack_vob_num ] -> num_cells ) ;

					assert (( play_stack [ stack_vob_num ] ->
								cells [ stack_cell_num ] ) != NULL ) ;
					play_stack_2 [ chapter ] -> cells [ curcell ] =
							calloc ( 1 , sizeof ( play_stack_cell ) ) ;
					memcpy ( play_stack_2 [ chapter ] -> 
								cells [ curcell ] ,
							play_stack [ stack_vob_num ] -> 
								cells [ stack_cell_num ] ,
							sizeof ( play_stack_cell ) ) ;
					play_stack [ stack_vob_num ] -> 
							cells [ stack_cell_num ] -> next = NULL ;
					play_stack_2 [ chapter ] -> num_cells ++ ;
					curcell ++ ;
				}
			}
		}
	}

	free_playback_stack ( play_stack , num_vobs ) ;

	/* attach play_stack to p */
	p -> play_stack_num_vobs = ptt_title -> num ;
	p -> play_stack = play_stack_2 ;
	p -> play_stack_title = title ;
	p -> play_stack_setup = 1 ;

	return IFO_OK ;
}

ifo_t * ifo_examine_title ( player_info_t * player_info , int title , int * fd )
{
	char		endpart [ 32 ] ;
	char		buf [ 256 ] ;
	ifo_t		* ifo ;

	if (( title > 0 )
	&& ( title < 100 ))
	{
		snprintf ( endpart , 32 , "/video_ts/vts_%02d_0.ifo" , title ) ;
	} else {
		strcpy ( endpart , "/video_ts/video_ts.ifo" ) ;
	}
	strcpy ( buf , player_info -> dvd_mountpoint ) ;
	strcat ( buf , endpart ) ;
	if ( ( (* fd) = open ( buf , O_RDONLY ) ) < 0 )
	{
		strcpy ( buf , player_info -> dvd_mountpoint ) ;
		string_uppercase ( endpart ) ;
		strcat ( buf , endpart ) ;
		if ( ( (* fd) = open ( buf , O_RDONLY ) ) < 0 )
		{
			return NULL ;
		}
	}

	if (( ifo = ifoOpen ( (* fd) , 0 ) ) == NULL )
	{
		close ( (* fd) ) ;
		(* fd) = -1 ;
		return NULL ;
	} else {
		return ifo ;
	}
}

static ifo_video_info_t * get_video_ptr ( player_info_t * p , int title )
{
	return (ifo_video_info_t *) ( get_ifo ( p , title ) -> tbl [ ID_MAT ] + IFO_OFFSET_VIDEO ) ;
}

static ifo_audio_t * get_audio_ptr ( player_info_t * p , int title , int stream )
{
	uint8_t		* table_ptr ;
	uint8_t 	* audio_ptr ;

	if ( ifoIsVTS ( get_ifo ( p , title ) ) ) 
	{
		return NULL ;
	}
	table_ptr = (uint8_t *) ( get_ifo ( p , title ) -> tbl [ ID_MAT ] + IFO_OFFSET_AUDIO ) ;
	if (( stream >= 0 )
	&& ( stream < ifoGetAudio ( table_ptr , & audio_ptr ) ))
	{
		/* I think IFO_AUDIO_LEN is off by one, it's 7, should be 8.
		 * That's where the 8 in the next line comes from. */
		return (ifo_audio_t *) ( audio_ptr + ( 8 * stream )) ;
	} else {
		return NULL ;
	}
}

static int get_first_time_map_table ( ifo_t * the_ifo , int * timeunit ,
					uint32_t ** start_ptr , uint32_t ** end_ptr )
{
	uint32_t		* ptr ;
	uint32_t		* end_of_table_ptr ;
	uint32_t		* safe_end_of_table_ptr ;
	ifo_hdr_t		* header ;
	ifoq_tmt_hdr_t	tmt_header ;
	uint16_t		num ;
	uint32_t		len , tmt_header_num ;


	if ( the_ifo == NULL ) 
	{
		return IFO_UNKNOWN_TITLE ;
	}

	ptr = (uint32_t *) ( the_ifo -> tbl [ ID_TMAPT ] ) ;
	if ( ptr == NULL ) 
	{
		return IFO_NO_TABLE ;
	}

	header = (ifo_hdr_t *) ptr ;

	num = be2me_16 ( header -> num ) ;
	len = be2me_32 ( header -> len ) ;
	if (( num == 0 )
	|| ( len == 0 )
	|| ( len > ( 1 << 20 ))) 		/* a simple test to see if len is */
	{								/* stupidly big.. */
		return IFO_SILLY_DATA ;
	}

	end_of_table_ptr = (uint32_t *) & ( ((u_char *) ptr ) [ len ] ) ;
	safe_end_of_table_ptr = end_of_table_ptr - 2 ;

	/* printf ( "len: %d   num: %d\n" , len , num ) ; */
	ptr += IFOQ_HDR_LEN ;
	ptr += num ;
	if ( ptr >= safe_end_of_table_ptr )
	{
		return IFO_SILLY_DATA ;
	}

	/* for ( i = 0 ; i < num ; i ++ )
	{
		printf ("%d start 0x%x\n", i, (uint32_t) be2me_32 ( (uint32_t) (* ptr ) ) ) ;
		ptr ++ ;
	} */

	/* byteswap the time unit header */
	tmt_header_num = be2me_32 ( (uint32_t) (* ptr) ) ;
	memset ( & tmt_header , 0 , sizeof ( ifoq_tmt_hdr_t ) ) ;
	memcpy ( & tmt_header , & tmt_header_num , sizeof ( uint32_t ) ) ;
	(* timeunit) = tmt_header . tu ;
	if ( (* timeunit) < 1 )
	{
		return IFO_SILLY_DATA ;
	}

	ptr += IFOQ_TMT_HDR_LEN ;

	if ( ptr >= safe_end_of_table_ptr )
	{
		return IFO_SILLY_DATA ;
	}

	(* start_ptr) = ptr ;
	(* end_ptr) = end_of_table_ptr ;

	return IFO_OK ;
}

static int playback_stack_check ( player_info_t * p , int title )
{
	if ( ! p -> ifo_available )
	{
		return IFO_NOT_AVAILABLE ;
	}
	if (( ! p -> play_stack_setup )
	|| ( p -> play_stack_title != title ))
	{
		p -> play_stack_accelerator_cell =
			p -> play_stack_accelerator_vob = -1 ;
		return compute_playback_stack ( p , title ) ;
	}
	return IFO_OK ;
}

