/*
 * Minidxr2 - Linux DVD player. Mostly (C) Jack Whitham, 2001
 *
 */
/*
 * output.c
 *
 * Provides the interface layer between the player and the decoder,
 * allowing things other than the output to be used to play DVDs.
 *
 */

#include <string.h>

#include "player.h"
#include "output.h"


int minidxr2_version_major = 2 ;
int minidxr2_version_minor = 1 ;

static void register_write ( output_plugin * o , int size ) ;

output_plugin * output_setup ( void * p , 
			output_plugin_fn out_fn , char ** err )
{
	output_plugin		* out ;

	/* copy the output_plugin */
	out = calloc ( 1 , sizeof ( output_plugin ) ) ;

	if ( ! ( (* out_fn) ( out ) ) )
	{
		(* err) = "Output plugin function returned zero.\n" ;
		free ( out ) ;
		return NULL ;
	}

	out -> css = calloc ( 1 , sizeof ( css_manager ) ) ;
	out -> output_mutex = calloc ( 1 , sizeof ( pthread_mutex_t ) ) ;
	pthread_mutex_init ( out -> output_mutex , NULL ) ;

	if ( out -> do_initialise != NULL )
	{
		out -> state_data =
				(* out -> do_initialise) ( p , err ) ;
	} else {
		(* err) = NULL ;
		out -> state_data = NULL ;
	}

	return out ;
}

char * output_deinitialise ( output_plugin * o )
{
	char *	err ;

	pthread_mutex_destroy ( o -> output_mutex ) ;
	free ( o -> output_mutex ) ;

	if ( o -> do_deinitialise != NULL )
	{
		err = (* o -> do_deinitialise) ( o -> state_data ) ;
	} else {
		err = NULL ;
	}
	free ( o ) ;

	return err ;
}


static int try_call_0_args ( output_plugin * o , output_control_fn_0_args proc )
{
	int		rc ;

#ifdef OUTPUT_DEBUG
	printf ( "output call 0 to 0x%x\n" , (int) proc ) ;
#endif
	pthread_mutex_lock ( o -> output_mutex ) ;
	if ( proc != NULL )
	{
		rc = (* proc) ( o -> state_data ) ;
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return rc ;
	} else {
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return OUTPUT_NOT_AVAILABLE ;
	}
}

static int try_call_1_args ( output_plugin * o , output_control_fn_1_args proc , int param )
{
	int		rc ;

#ifdef OUTPUT_DEBUG
	printf ( "output call 1 to 0x%x ( %d ) \n" , (int) proc , param ) ;
#endif
	pthread_mutex_lock ( o -> output_mutex ) ;
	if ( proc != NULL )
	{
		rc = (* proc) ( o -> state_data , param ) ;
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return rc ;
	} else {
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return OUTPUT_NOT_AVAILABLE ;
	}
}

static int try_call_2_args ( output_plugin * o , output_control_fn_2_args proc , int param1 , int param2 )
{
	int		rc ;

#ifdef OUTPUT_DEBUG
	printf ( "output call 2 to 0x%x ( %d , %d ) \n" , (int) proc , param1 , param2 ) ;
#endif
	pthread_mutex_lock ( o -> output_mutex ) ;
	if ( proc != NULL )
	{
		rc = (* proc) ( o -> state_data , param1 , param2 ) ;
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return rc ;
	} else {
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return OUTPUT_NOT_AVAILABLE ;
	}
}


int output_decryption_switch ( output_plugin * o , int param )
{
	return try_call_1_args ( o , o -> decryption_switch , param ) ;
}

int output_fast_forwards ( output_plugin * o , int param )
{
	return try_call_1_args ( o , o -> fast_forwards , param ) ;
}

int output_get_playmode ( output_plugin * o )
{
	return try_call_0_args ( o , o -> get_playmode ) ;
}

int output_mute ( output_plugin * o , int param )
{
	return try_call_1_args ( o , o -> mute , param ) ;
}

int output_reset ( output_plugin * o )
{
	return try_call_0_args ( o , o -> reset ) ;
}

int output_select_audio ( output_plugin * o , int param1 , int param2 )
{
	return try_call_2_args ( o , o -> select_audio , param1 , param2 ) ;
}

int output_select_subpicture ( output_plugin * o , int param )
{
	return try_call_1_args ( o , o -> select_subpicture , param ) ;
}

int output_set_subpicture_palette ( output_plugin * o , clut_t * pal )
{
	return try_call_1_args ( o , o -> set_subpicture_palette , (int) pal ) ;
} 

int output_set_audio_freq ( output_plugin * o , int param )
{
	return try_call_1_args ( o , o -> set_audio_freq , param ) ;
}

int output_set_audio_width ( output_plugin * o , int param )
{
	return try_call_1_args ( o , o -> set_audio_width , param ) ;
}

int output_set_playmode ( output_plugin * o , int param )
{
	return try_call_1_args ( o , o -> set_playmode , param ) ;
}

int output_subpicture ( output_plugin * o , int param )
{
	return try_call_1_args ( o , o -> subpicture , param ) ;
}

/* special controls */
int output_write ( output_plugin * o , char * data , int param )
{
	int		rc ;

#ifdef OUTPUT_DEBUG
	printf ( "output " __FUNCTION__ "( .. , %d )\n" , param ) ;
#endif
	pthread_mutex_lock ( o -> output_mutex ) ;
	if ( o -> do_write != NULL )
	{
		register_write ( o , param ) ;
		rc = (* o -> do_write) ( o -> state_data , data , param ) ;
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return rc ;
	} else {
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return OUTPUT_NOT_AVAILABLE ;
	}
}

int output_ioctl ( output_plugin * o , int num , void * data )
{
	int		rc ;

#ifdef OUTPUT_DEBUG
	printf ( "output " __FUNCTION__ "( %d , 0x%x )\n" , num , data ) ;
#endif
	pthread_mutex_lock ( o -> output_mutex ) ;
	if ( o -> do_ioctl != NULL )
	{
		rc = (* o -> do_ioctl) ( o -> state_data , num , data ) ;
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return rc ;
	} else {
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return OUTPUT_NOT_AVAILABLE ;
	}
}

int output_css_authenticate ( output_plugin * o , char * dvd_device ,
					char * err , int max_chars )
{
	int		rc ;

#ifdef OUTPUT_DEBUG
	printf ( "output " __FUNCTION__ "( %s )\n" , dvd_device ) ;
#endif
	pthread_mutex_lock ( o -> output_mutex ) ;
	if ( o -> do_css_authenticate != NULL )
	{
		rc = (* o -> do_css_authenticate) ( o -> css ,
				o -> state_data , dvd_device , err , max_chars ) ;
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return rc ;
	} else {
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return OUTPUT_NOT_AVAILABLE ;
	}
}

int output_css_finish ( output_plugin * o )
{
	int		rc ;

#ifdef OUTPUT_DEBUG
	printf ( "output " __FUNCTION__ "()\n" ) ;
#endif
	pthread_mutex_lock ( o -> output_mutex ) ;
	if ( o -> do_css_finish != NULL )
	{
		rc = (* o -> do_css_finish) ( o -> css ) ;
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return rc ;
	} else {
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return OUTPUT_NOT_AVAILABLE ;
	}
}

int output_css_do_disc_key ( output_plugin * o , char * err , int max_chars )
{
	int		rc ;

#ifdef OUTPUT_DEBUG
	printf ( "output " __FUNCTION__ "()\n" ) ;
#endif
	pthread_mutex_lock ( o -> output_mutex ) ;
	if ( o -> do_css_do_disc_key != NULL )
	{
		rc = (* o -> do_css_do_disc_key) ( o -> css , 
					o -> state_data , err , max_chars ) ;
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return rc ;
	} else {
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return OUTPUT_NOT_AVAILABLE ;
	}
}

int output_css_do_title_key ( output_plugin * o , char * err , int max_chars , int lba )
{
	int		rc ;

#ifdef OUTPUT_DEBUG
	printf ( "output " __FUNCTION__ "( %d ) \n" , lba ) ;
#endif
	pthread_mutex_lock ( o -> output_mutex ) ;
	if ( o -> do_css_do_title_key != NULL )
	{
		rc = (* o -> do_css_do_title_key) ( o -> css , 
					o -> state_data , err , max_chars , lba ) ;
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return rc ;
	} else {
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return OUTPUT_NOT_AVAILABLE ;
	}
}

char * output_css_do_reset ( output_plugin * o , char * dvd_device )
{
	char		* err ;

	pthread_mutex_lock ( o -> output_mutex ) ;
	if ( o -> do_css_reset != NULL )
	{
		err = (* o -> do_css_reset) ( dvd_device ) ;
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return err ;
	} else {
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return "CSS reset function not available\n" ;
	}
}

int output_set_video_type ( output_plugin * o , int type )
{
	return try_call_1_args ( o , o -> set_video_type , type ) ;
}

int output_set_scaling_mode ( output_plugin * o , int mode )
{
	return try_call_1_args ( o , o -> set_scaling_mode , mode ) ;
}

int output_set_view_mode ( output_plugin * o , int mode , geom_t geom )
{
	int		rc ;

	pthread_mutex_lock ( o -> output_mutex ) ;
	if ( o -> do_set_view_mode != NULL )
	{
		rc = (* o -> do_set_view_mode) ( o -> state_data , mode , geom ) ;
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return rc ;
	} else {
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return 0 ;
	}
}

int output_initialise_overlay ( output_plugin * o , output_extra_info * ck )
{
	int		rc ;

	pthread_mutex_lock ( o -> output_mutex ) ;
	if ( o -> do_initialise_overlay != NULL )
	{
		rc = (* o -> do_initialise_overlay) ( o -> state_data , ck ) ;
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return rc ;
	} else {
		pthread_mutex_unlock ( o -> output_mutex ) ;
		return 0 ;
	}
}

/* Translation table functions */
/* These are currently used by the dxr2 code to translate internal
 * constants (representing commands to the player) into ioctl numbers
 * that the dxr2 api can understand. */

int output_int_to_ext ( output_xlat * xlat , int internal )
{
	int				i = 0 ;

	while ( xlat [ i ] . internal_value != OUTPUT_END )
	{
		if ( xlat [ i ] . internal_value == internal )
		{
			return xlat [ i ] . external_value ;
		}
		i ++ ;
	}
	return OUTPUT_NOT_AVAILABLE ;
}

int output_ext_to_int ( output_xlat * xlat , int external )
{
	int				i = 0 ;

	while ( xlat [ i ] . internal_value != OUTPUT_END )
	{
		if ( xlat [ i ] . external_value == external )
		{
			return xlat [ i ] . internal_value ;
		}
		i ++ ;
	}
	return OUTPUT_NOT_AVAILABLE ;
}

void output_if_valid_int_then_set ( output_xlat * xlat , int * set_me , int int_value )
{
	int			rc ;

	rc = output_int_to_ext ( xlat , int_value ) ;
	if ( rc != OUTPUT_NOT_AVAILABLE )
	{
		(* set_me) = rc ;
	}
}



static void register_write ( output_plugin * o , int size )
{
	time_t		current_time ;

	current_time = time ( NULL ) ;
	if ( current_time != o -> bitrate_records_start )
	{
		memmove ( o -> bitrate_records_total ,
					& o -> bitrate_records_total [ 1 ] ,
					sizeof ( int ) * ( NUM_BITRATE_RECORDS - 1 ) ) ;
		o -> bitrate_records_total [ NUM_BITRATE_RECORDS - 1 ] = 0 ;
		o -> bitrate_records_start = current_time ;
	}
	o -> bitrate_records_total [ NUM_BITRATE_RECORDS - 1 ] += size ;
}

/* Actually returns the average kbitrate over the
 * last NUM_BITRATE_RECORDS - 1 seconds during which playback
 * was in progress. */
int output_get_bitrate ( output_plugin * o )
{
	int			i ;
	int			total = 0 ;

	pthread_mutex_lock ( o -> output_mutex ) ;
	for ( i = 0 ; i < ( NUM_BITRATE_RECORDS - 1 ) ; i ++ )
	{
		total += ( o -> bitrate_records_total [ i ] ) / 
					(( NUM_BITRATE_RECORDS - 1 ) * ( 1024 / 8 )) ;
	}
	pthread_mutex_unlock ( o -> output_mutex ) ;

	return total ;
}

